
from __future__ import division
import scipy as s

from .nodes import *
from .distributions import *


"""
This module is used to define nodes that are infered using the variational bayes framework.

All variational nodes share the property of having a lower bound associated with it.
We divide variational nodes into Constant and Unobserved. 
The constant variational nodes are fixed and have no parameters or expectations
The unobserved variational nodes are learnt and have an associated P and Q distribution with corresponding expectations and parameters
"""

###########################################
## General classes for variational nodes ##
###########################################

class Variational_Node(Node):
    """
    Abstract class for a variational node in a Bayesian probabilistic model.
    Variational nodes can be observed (constant) or unobserved
    """
    def __init__(self, dim):
        Node.__init__(self,dim)

    def calculateELBO(self):
        # General method to calculate the ELBO of the node
        return 0.

###################################################################
## General classes for observed and unobserved variational nodes ##
###################################################################

class Constant_Variational_Node(Variational_Node,Constant_Node):
    """
    Abstract class for an observed/constant variational node in a Bayesian probabilistic model.
    """
    def __init__(self, dim, value):
        # SHOULD WE ALSO INITIALISE VARIATIONAL_NODE ..?
        Constant_Node.__init__(self, dim, value)


class Unobserved_Variational_Node(Variational_Node):
    """
    Abstract class for an unobserved variational node in a Bayesian probabilistic model.
    Unobserved variational nodes contain a prior P(X) and a variational Q(X) distribution,
    which will be stored as instances of Distribution() attributes .P and .Q, respectively.
    The distributions are in turn composed of parameters and expectations
    """
    def __init__(self, dim):
        Variational_Node.__init__(self, dim)
        self.P = None
        self.Q = None
    def updateExpectations(self, dist="Q"):
        # Method to update expectations of the node
        if dist == "Q": self.Q.updateExpectations()

    def getExpectation(self, dist="Q"):
        # Method to get the first moment (expectation) of the node
        if dist == "Q": expectations = self.Q.getExpectations()
        elif dist == "P": expectations = self.P.getExpectations()
        return expectations["E"]

    def getExpectations(self, dist="Q"):
        # Method to get all relevant moments of the node
        if dist == "Q": expectations = self.Q.getExpectations()
        elif dist == "P": expectations = self.P.getExpectations()
        return expectations

    def getParameters(self, dist="Q"):
        # Method to get all parameters of the node
        if dist == "Q": params = self.Q.getParameters()
        elif dist == "P": params = self.P.getParameters()
        return params

    def removeFactors(self, idx, axis=None):
        # Method to remove entire factors from the nodes
        if hasattr(self,"factors_axis"): axis = self.factors_axis
        if hasattr(self,"covariates"): self.covariates = self.covariates[np.isin(s.arange(len(self.covariates)), idx, invert=True)]
        if axis is not None:
            self.P.removeDimensions(axis=axis, idx=idx)
            self.Q.removeDimensions(axis=axis, idx=idx)
            self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))

#######################################################
## Specific classes for unobserved variational nodes ##
#######################################################

class UnivariateGaussian_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where P(.) and Q(.)
    are both univariate Gaussian distributions.
    """
    def __init__(self, dim, pmean, pvar, qmean, qvar, qE=None, qE2=None):
	    # dim (2d tuple): dimensionality of the node
	    # pmean (nd array): the mean parameter of the P distribution
	    # qmean (nd array): the mean parameter of the Q distribution
	    # pvar (nd array): the variance parameter of the P distribution
	    # qvar (nd array): the variance parameter of the Q distribution
	    # qE (nd array): the initial first moment of the Q distribution
        Unobserved_Variational_Node.__init__(self, dim)
        # Initialise the P and Q distributions
        self.P = UnivariateGaussian(dim=dim, mean=pmean, var=pvar)
        self.Q = UnivariateGaussian(dim=dim, mean=qmean, var=qvar, E=qE, E2=qE2)

class MultivariateGaussian_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where P(.) and Q(.)
    are both multivariate Gaussian distributions.
    """
    def __init__(self, dim, pmean, pcov, qmean, qcov, qE=None):
        # dim (2d tuple): dimensionality of the node
        # pmean (nd array): the mean parameter of the P distribution
        # pcov (nd array): the covariance parameter of the P distribution
        # qmean (nd array): the mean parameter of the Q distribution
        # qcov (nd array): the covariance parameter of the Q distribution
        # qE (nd array): the initial first moment of the Q distribution
        Unobserved_Variational_Node.__init__(self, dim)
        
        # Initialise the P and Q distributions
        self.P = MultivariateGaussian(dim=dim, mean=pmean, cov=pcov)
        self.Q = MultivariateGaussian(dim=dim, mean=qmean, cov=qcov, E=qE)

class Gamma_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where P(x) and Q(x) are both gamma distributions
    """
    def __init__(self, dim, pa, pb, qa, qb, qE=None):
	    # dim (2d tuple): dimensionality of the node
	    # pa (nd array): the 'a' parameter of the P distribution
	    # qa (nd array): the 'b' parameter of the P distribution
	    # qa (nd array): the 'a' parameter of the Q distribution
	    # qb (nd array): the 'b' parameter of the Q distribution
	    # qE (nd array): the initial expectation of the Q distribution
        Unobserved_Variational_Node.__init__(self,dim)

        # Initialise the distributions
        self.P = Gamma(dim=dim, a=pa, b=pb)
        self.Q = Gamma(dim=dim, a=qa, b=qb, E=qE)

class Bernoulli_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where P(.) and Q(.)
    are both bernoulli distributions.
    """
    def __init__(self, dim, ptheta, qtheta, qE=None):
	    # dim (2d tuple): dimensionality of the node
	    # ptheta (nd array): the 'theta' parameter of the P distribution
	    # qtheta (nd array): the 'theta' parameter of the Q distribution
	    # qE (nd array): initial first moment of the Q distribution
        Unobserved_Variational_Node.__init__(self,dim)

        # Initialise the distributions
        self.P = Bernoulli(dim=dim, theta=ptheta)
        self.Q = Bernoulli(dim=dim, theta=qtheta, E=qE)

class BernoulliGaussian_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where P(.) and Q(.)
    are joint gaussian-bernoulli distributions (see paper  Spike and Slab Variational Inference for
    Multi-Task and Multiple Kernel Learning by Titsias and Gredilla)
    """
    def __init__(self, dim,
        pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta,
        qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0=None, qEW_S1=None, qES=None):
	    # dim (2d tuple): dimensionality of the node
        # pmean (nd array): the mean parameter of the P distribution
        # pvar (nd array): the var parameter of the P distribution
	    # ptheta (nd array): the theta parameter of the P distribution
        # qmean (nd array): the mean parameter of the Q distribution
        # qvar (nd array): the var parameter of the Q distribution
	    # qtheta (nd array): the theta parameter of the Q distribution
        Unobserved_Variational_Node.__init__(self,dim)

        # Initialise the P and Q distributions
        self.P = BernoulliGaussian(dim=dim, theta=ptheta, mean_S0=pmean_S0, var_S0=pvar_S0, mean_S1=pmean_S1, var_S1=pvar_S1)
        self.Q = BernoulliGaussian(dim=dim, theta=qtheta, mean_S0=qmean_S0, var_S0=qvar_S0, mean_S1=qmean_S1, var_S1=qvar_S1, EW_S0=qEW_S0, EW_S1=qEW_S1, ES=qES)

class Beta_Unobserved_Variational_Node(Unobserved_Variational_Node):
    """
    Abstract class for a variational node where both P(.) and Q(.) are beta
    distributions
    """
    def __init__(self, dim, pa, pb, qa, qb, qE=None):
        # dim (2d tuple): dimensionality of the node
        # pa (nd array): the 'a' parameter of the P distribution
        # qa (nd array): the 'b' parameter of the P distribution
        # qa (nd array): the 'a' parameter of the Q distribution
        # qb (nd array): the 'b' parameter of the Q distribution
        # qE (nd array): the initial expectation of the Q distribution
        super(Beta_Unobserved_Variational_Node, self).__init__(dim)

        # Initialise P and Q distributions
        self.P = Beta(dim, pa, pb)
        self.Q = Beta(dim, qa, qb, qE)



import pandas as pd
from mofapy.core.entry_point import entry_point

# Load the data 
# Be careful to use the right delimiter, and make sure that you use the right arguments from pandas.read_csv to load the row names and column names, if appropriate.
M = 3 # Number of views
data =  [None]*M
data[0] = pd.read_csv("(...)/view_0.txt", delimiter=" ").astype(pd.np.float32)
data[1] = pd.read_csv("(...).txt", delimiter=" ").astype(pd.np.float32)
data[2] = pd.read_csv("(...).txt", delimiter=" ").astype(pd.np.float32)

# Initialise entry point
ep = entry_point()

# Set data
ep.set_data(data)

## Set model options ##
# factors: number of factors. By default, the model does not automatically learn the number of factors. 
# 	If you want the model to do this (based on a minimum variance explained criteria), set `TrainOptions$dropFactorThreshold` to a non-zero value.
# likelihoods: list with the likelihood for each view. Usually we recommend: 
#	- gaussian for continuous data
# 	- bernoulli for binary data
#  	- poisson for count data
# 	If you are using gaussian likelihoods, we recommend centering the data (specified in data_options) and setting learnIntercept to False. 
# 	However, if you have non-gaussian likelihoods, learning an intercept factor is important
# sparsity: boolean indicating whether to use sparsity. 
# 	This is always recommended, as it will make the loadings more interpretable.
ep.set_model_options(factors=25, likelihoods=["gaussian","bernoulli","poisson"], sparsity=True)

## Set data options ##

# view_names: list with view names
# center_features: boolean indicating whether to center the features to zero mean. 
# 	This only works for gaussian data. Default is TRUE.
# scale_views: boolean indicating whether to scale views to have the same unit variance. 
# 	As long as the scale differences between the data sets is not too high, this is not required. Default is False.
# RemoveIncompleteSamples: boolean indicating whether to remove samples that are not profiled in all omics. 
# 	We recommend this only for testing, as the model can cope with samples having missing assays. Default is False.
ep.set_data_options(center_features=True, scale_views=False, RemoveIncompleteSamples=False)

# Parse the data (optionally center or scale, do some QC, etc.)
ep.parse_data()

# Define training options
# iter: numeric value indicating the maximum number of iterations. 
# 	Default is 1000, we recommend setting this to a large value and using the 'tolerance' as convergence criteria.
# tolerance: numeric value indicating the convergence threshold based on the change in Evidence Lower Bound (deltaELBO). 
# 	For quick exploration we recommend this to be around 1.0, and for a thorough training we recommend a value of 0.01. Default is 0.1
# dropR2: numeric hyperparamter to automatically learn the number of factors. 
# 	It indicates the threshold on fraction of variance explained to consider a factor inactive and automatically drop it from the model during training. 
# 	For example, a value of 0.01 implies that factors explaining less than 1% of variance (in each view) will be dropped. 
# 	Default is 0, which implies that only factors that explain no variance at all will be removed
# elbofreq: frequency of computation of the ELBO. It is useful to assess convergence, but it slows down the training.
# verbose: boolean indicating whether to generate a verbose output.
# seed: random seed. If None, it is sampled randomly
ep.set_train_options(iter=1000, tolerance=0.01, dropR2=0.00, elbofreq=1, verbose=False, seed=2018)

# Define prior distributions
ep.define_priors()

# Define initialisations of variational distributions
ep.define_init()

# Parse intercept factor
ep.parse_intercept()

# Train the model
ep.train_model()

# Save the model
outfile="/Users/mofa/test.hdf5"
ep.save_model(outfile)

# This is the end of the python pipeline.
# Now you are ready to switch to the R framework for downstream analysis:
# (in R) model <- loadModel("/Users/mofa/test.hdf5")



from __future__ import division
from time import sleep
from copy import deepcopy

import numpy as np
import numpy.ma as ma
import os
import h5py
import sys

"""
Module to define util functions

TO-DO: 
- Create a proper class for saving the models
- Create a proper class for loading and parsing the data
- Move the util math functions into a math.py
"""


def removeIncompleteSamples(data):
    """ Method to remove samples with missing views 

    PARAMETERS
    ----------
    data: list  
    """
    print("\nRemoving incomplete samples...")

    M = len(data)
    N = data[0].shape[0]
    samples_to_remove = []
    for n in range(N):
        for m in range(M):
            if np.all(np.isnan(data[m][n,:])):
                samples_to_remove.append(n)
                break

    if len(samples_to_remove) > 0:
        print("A total of " + str(len(samples_to_remove)) + " sample(s) have at least a missing view and will be removed")

    data_filt = [None]*M
    samples_to_keep = np.setdiff1d(range(N),samples_to_remove)
    for m in range(M):
        data_filt[m] = data[m][samples_to_keep,:]

    return data_filt

def maskData(data, data_opts):
    """ Method to mask values of the data, 
    It is mainly to test missing values and to evaluate imputation
    
    PARAMETERS
    ----------
    data_opts: dic
    
    """
    print("Not functional")
    sys.stdout.flush()
    exit()

    print("Masking data with the following options:")
    print("at random:")
    print(data_opts['maskAtRandom'])
    print("full cases:")
    print(data_opts['maskNSamples'])


    for m in range(len(data)):

        # Mask values at random
        D = data[m].shape[1]
        N = data[m].shape[0]
        p2Mask = data_opts['maskAtRandom'][m]
        if p2Mask != 0:
            idxMask = np.zeros(N*D)
            idxMask[:int(round(N*D*p2Mask))] = 1
            np.random.shuffle(idxMask)
            idxMask = np.reshape(idxMask, [N, D])
            data[m] = data[m].mask(idxMask==1)

        # Mask samples in a complete view
        Nsamples2Mask = data_opts['maskNSamples'][m]
        if Nsamples2Mask != 0:
            idxMask = np.random.choice(N, size=Nsamples2Mask, replace = False)
            # idxMask = np.arange(Nsamples2Mask)
            # print idxMask
            tmp = data[m].copy()
            tmp.ix[idxMask,:] = pd.np.nan
            data[m] = tmp

    return data

def parseData(data, data_opts):
    """ Method to do parse the data
    
    TO-DO: CHECK IF ANY SAMPLE HAS MISSING VALUES IN ALL VIEWS 

    PARAMETERS
    ----------
    data: list of numpy arrays or pandas dataframes
    """
    M = len(data)
    parsed_data = deepcopy(data)
    for m in range(M):
        # Convert to float32
        # parsed_data[m] = parsed_data[m].astype(pd.np.float32)
        parsed_data[m] = parsed_data[m].astype(np.float32)

        # For some reason, reticulate stores missing values in integer matrices as -2147483648
        parsed_data[m][parsed_data[m] == -2147483648] = np.nan
        # Center the features
        if data_opts['center_features'][m]:
            print("Centering features for view " + data_opts["view_names"][m] + "...")
            parsed_data[m] = parsed_data[m] - np.nanmean(parsed_data[m],axis=0)

        # Scale the views to unit variance
        if data_opts['scale_views'][m]:
            print("Scaling view " + data_opts["view_names"][m] + " to unit variance...")
            parsed_data[m] = parsed_data[m] / np.nanstd(parsed_data[m])

        # Scale the features to unit variance
        if data_opts['scale_features'][m]:
            print("Scaling features for view " + data_opts["view_names"][m] + " to unit variance...")
            parsed_data[m] = parsed_data[m] / np.nanstd(parsed_data[m], axis=0, )

    print("\nAfter parsing the data:")
    for m in range(M): print("view %s has %d samples and %d features..." % (data_opts["view_names"][m], parsed_data[m].shape[0], parsed_data[m].shape[1]))

    return parsed_data

def qcData(data):
    """ Method to do quality control on the data
    
    TO-DO: CHECK IF ANY SAMPLE HAS MISSING VALUES IN ALL VIEWS 

    PARAMETERS
    ----------
    data: list of numpy arrays or pandas dataframes
    """

    M = len(data)

    # Check that the dimensions match
    if len(set([data[m].shape[0] for m in range(M)])) != 1:
        if all([data[m].shape[1] for m in range(M)]):
            print("\nWarning: columns seem to be the shared axis, transposing the data...")
            for m in range(M): data[m] = data[m].T
        else:
            print("\nError: Dimensionalities do not match, aborting. Make sure that either columns or rows are shared!")
            sys.stdout.flush()
            exit()


    # Sanity checks on the data
    print ("\n" +"#"*46)
    print("## Doing sanity checks and parsing the data ##")
    print ("#"*46 + "\n")
    for m in range(M):

        # Detect features with complete missing values
        nas = np.isnan(data[m]).mean(axis=0)
        if np.any(nas==1.):
            print("Error: %d features(s) on view %d have missing values in all samples, please remove them before running the model." % ( (nas==1.).sum(), m) )
            sys.stdout.flush()
            exit()

        # Detect features with no variance
        var = data[m].std(axis=0) 
        if np.any(var==0.):
            print("Error: %d features(s) on view %d have zero variance. Please, remove lowly variable features for each omic separately, they can cause numerical issues." % ( (var==0.).sum(),m) )
            sys.stdout.flush()
            exit()

    return data

def dotd(A, B, out=None):
    """Diagonal of :math:`\mathrm A\mathrm B^\intercal`.
    If ``A`` is :math:`n\times p` and ``B`` is :math:`p\times n`, it is done in :math:`O(pn)`.
    Args:
        A (array_like): Left matrix.
        B (array_like): Right matrix.
        out (:class:`numpy.ndarray`, optional): copy result to.
    Returns:
        :class:`numpy.ndarray`: Resulting diagonal.
    """
    A = ma.asarray(A, float)
    B = ma.asarray(B, float)
    if A.ndim == 1 and B.ndim == 1:
        if out is None:
            return ma.dot(A, B)
        return ma.dot(A, B, out)

    if out is None:
        out = ma.empty((A.shape[0], ), float)

    out[:] = ma.sum(A * B.T, axis=1)
    return out

def nans(shape, dtype=float):
    """ Method to create an array filled with missing values """
    a = np.empty(shape, dtype)
    a.fill(np.nan)
    return a

def corr(A,B):
    """ Method to efficiently compute correlation coefficients between two matrices 
    
    PARMETERS
    ---------
    A: np array
    B: np array
    """

    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - A.mean(1)[:,None]
    B_mB = B - B.mean(1)[:,None]

    # Sum of squares across rows
    ssA = (A_mA**2).sum(1);
    ssB = (B_mB**2).sum(1);

    # Finally get corr coeff
    return np.dot(A_mA,B_mB.T)/np.sqrt(np.dot(ssA[:,None],ssB[None]))

# NOT HERE
def logdet(X):
    return np.log(np.linalg.det(X))
    # UC = np.linalg.cholesky(X)
    # return 2*sum(np.log(np.diag(UC)))


# NOT HERE
def ddot(d, mtx, left=True):
    """Multiply a full matrix by a diagonal matrix.
    This function should always be faster than dot.

    Input:
      d -- 1D (N,) array (contains the diagonal elements)
      mtx -- 2D (N,N) array
      left: is the diagonal matrix on the left or on the right of the product?

    Output:
      ddot(d, mts, left=True) == dot(diag(d), mtx)
      ddot(d, mts, left=False) == dot(mtx, diag(d))
    """
    if left:
        return (d*mtx.T).T
    else:
        return d*mtx


def saveParameters(model, hdf5, view_names=None):
    """ Method to save the parameters of the model in an hdf5 file
    
    PARAMETERS
    ----------
    model: a BayesNet instance
    hdf5: 
    view_names
    """
    
    # Get nodes from the model
    nodes = model.getNodes()

    # Create groups
    param_grp = hdf5.create_group("parameters")

    # Iterate over nodes
    for node in nodes:

        # Collect node parameters
        parameters = nodes[node].getParameters()

        # Create node subgroup
        node_subgrp = param_grp.create_group(node)

        # Multi-view nodes
        if type(parameters) == list:
            # Loop through the views
            for m in range(len(parameters)):
                if view_names is not None:
                    tmp = view_names[m]
                else:
                    tmp = "%d" % m
                # Create subsubgroup for the view
                view_subgrp = node_subgrp.create_group(tmp)
                # Loop through the parameters of the view
                if parameters[m] is not None:
                    # Variational nodes
                    if type(parameters[m]) == dict:
                        for param_name in parameters[m].keys():
                            if parameters[m][param_name] is not None:
                                view_subgrp.create_dataset(param_name, data=parameters[m][param_name].T)
                    # Non-variational nodes (no distributions)
                    elif type(parameters[m]) == np.ndarray:
                           view_subgrp.create_dataset("value", data=parameters[m].T)

        # Single-view nodes
        else:
            for param_name in parameters.keys():
                node_subgrp.create_dataset("%s" % (param_name), data=parameters[param_name].T)
    pass

def saveExpectations(model, hdf5, view_names=None):
    """ Method to save the expectations of the model in an hdf5 file
    
    PARAMETERS
    ----------
    model: a BayesNet instance
    hdf5: 
    view_names:
    """
    # Get nodes from the model
    nodes = model.getNodes()

    exp_grp = hdf5.create_group("expectations")

    # Iterate over nodes
    for node in nodes:

        # Collect node expectations
        expectations = nodes[node].getExpectations()


        # Multi-view nodes
        if type(expectations) == list:

            # Create subgroup for the node
            node_subgrp = exp_grp.create_group(node)

            # Iterate over views
            for m in range(len(expectations)):
                if view_names is not None:
                    view = view_names[m]
                else:
                    view = "%d" % m

                # Collect expectations
                exp = expectations[m]["E"]
                if exp  is not None:
                    if type(exp) == ma.core.MaskedArray:
                        tmp = ma.filled(exp, fill_value=np.nan)
                        node_subgrp.create_dataset(view, data=tmp.T)
                    else:
                        node_subgrp.create_dataset(view, data=exp.T)

        # Single-view nodes
        else:
            exp_grp.create_dataset(node, data=expectations["E"].T)

def saveTrainingStats(model, hdf5):
    """ Method to save the training statistics in an hdf5 file
    
    PARAMETERS
    ----------
    model: a BayesNet instance
    hdf5: 
    """
    stats = model.getTrainingStats()
    stats_grp = hdf5.create_group("training_stats")
    stats_grp.create_dataset("activeK", data=stats["activeK"])
    stats_grp.create_dataset("elbo", data=stats["elbo"])
    stats_grp.create_dataset("elbo_terms", data=stats["elbo_terms"].T)
    stats_grp['elbo_terms'].attrs['colnames'] = [a.encode('utf8') for a in stats["elbo_terms"].columns.values]

def saveTrainingOpts(opts, hdf5):
    """ Method to save the training options in an hdf5 file
    
    PARAMETERS
    ----------
    opts:
    hdf5: 
    """
    # Remove dictionaries from the options
    for k,v in opts.copy().items():
        if type(v)==dict:
            for k1,v1 in v.items():
                opts[str(k)+"_"+str(k1)] = v1
            opts.pop(k)

    if 'schedule' in opts.keys():
        del opts['schedule']

    # Create HDF5 data set
    hdf5.create_dataset("training_opts", data=np.array(list(opts.values()), dtype=np.float))
    hdf5['training_opts'].attrs['names'] = np.asarray(list(opts.keys())).astype('S')

def saveModelOpts(opts, hdf5):
    """ Method to save the model options in an hdf5 file
    
    PARAMETERS
    ----------
    opts: model options
    hdf5: h5py.File instance
    """
    opts_interest = ["learnIntercept","likelihoods","sparsity_bool","factors"]
    opts = dict((k, opts[k]) for k in opts_interest)
    opts["sparsity"] = opts.pop("sparsity_bool")
    opts["numFactors"] = opts.pop("factors")
    grp = hdf5.create_group('model_opts')
    for k,v in opts.items():
        grp.create_dataset(k, data=np.asarray(v).astype('S'))
    grp[k].attrs['names'] = np.asarray(list(opts.keys())).astype('S')

def saveTrainingData(model, hdf5, data, view_names=None, sample_names=None, feature_names=None, likelihoods=None):
    """ Method to save the training data in an hdf5 file
    
    PARAMETERS
    ----------
    model: a BayesNet instance
    hdf5: h5py.File instance
    view_names: list with view names as characters. If None, no view names will be saved
    sample_names: list with sample names as characters. If None, no sample names will be saved
    feature_names: list with feature names as characters. If None, no feature names will be saved
    """
    # data = model.getTrainingData()

    # Make sure that missing values are masked
    # for i in range(len(data)):
    #     data[i].data[data[i].mask] = np.nan

    data_grp = hdf5.create_group("data")
    samples_grp = hdf5.create_group("samples")
    features_grp = hdf5.create_group("features")
    intercept_grp = hdf5.create_group("intercept")

    if sample_names is not None:
        samples_grp.create_dataset("samples", data=np.array(sample_names, dtype='S50'))

    # if likelihoods is not None:
    #     data_grp.attrs['likelihood'] = np.array(likelihoods, dtype='S50')

    for m in range(len(data)):
        view = view_names[m] if view_names is not None else str(m)
        data_grp.create_dataset(view, data=data[m].T)
        intercept_grp.create_dataset(view, data=np.nanmean(data[m],axis=0))
        # if likelihoods[m] is "gaussian":
        # else:
            # intercept_grp.create_dataset(view, data=model.getNodes()["Y"].nodes[m].means)
        if feature_names is not None:
            features_grp.create_dataset(view, data=np.array(feature_names[m], dtype='S50'))
        

def saveModel(model, data, outfile, train_opts, model_opts, view_names=None, sample_names=None, feature_names=None):
    """ Method to save the model in an hdf5 file
    
    PARAMETERS
    ----------
    TO-FILL....
    """

    # QC checks
    assert model.trained == True, "Model is not trained yet"
    if view_names is not None:
        assert len(np.unique(view_names)) == len(view_names), 'View names must be unique'
    if sample_names is not None:
        assert len(np.unique(sample_names)) == len(sample_names), 'Sample names must be unique'
    if feature_names is not None:
        for x in feature_names: assert len(np.unique(x)) == len(x), 'Feature names must be unique'

    # Create output directory
    if not os.path.isdir(os.path.dirname(outfile)):
        print("Output directory does not exist, creating it...")
        os.makedirs(os.path.dirname(outfile))

    # For some reason h5py orders the datasets alphabetically, so we have to sort the likelihoods accordingly
    idx = sorted(range(len(view_names)), key=lambda k: view_names[k])
    tmp = [model_opts["likelihoods"][idx[m]] for m in range(len(model_opts["likelihoods"]))]
    model_opts["likelihoods"] = tmp

    # Open HDF5 handler
    hdf5 = h5py.File(outfile,'w')

    # Save expectations
    saveExpectations(model,hdf5,view_names)

    # Save parameters
    # saveParameters(model,hdf5,view_names)

    # Save training statistics
    saveTrainingStats(model,hdf5)

    # Save training options
    saveTrainingOpts(train_opts,hdf5)

    # Save model options
    saveModelOpts(model_opts,hdf5)

    # Save training data
    saveTrainingData(model, hdf5, data, view_names, sample_names, feature_names, model_opts["likelihoods"])

    # Close HDF5 file
    hdf5.close()



"""
Module to simulate test data

To-do:
- Currently non-gaussian likelihoods have no noise
- Maybe integrate this into the corresponding Node classes with a Simulate() method?
- Fix binomial
"""

from __future__ import division
import scipy as s
import pandas as pd
import warnings
from scipy.stats import bernoulli, norm, gamma, uniform, poisson, binom
from random import sample

def sigmoid(X):
    return s.divide(1.,1.+s.exp(-X))

class Simulate(object):
    def __init__(self, M, N, D, K):
        """General method to Simulate from the generative model

        PARAMETERS
        ----------
        M (int): number of views
        N (int): number of samples
        D (list/tuple of length M): dimensionality of each view
        K (int): number of latent variables
        """

        # Sanity checks
        assert len(D) == M
        assert K < min(D)
        assert K < N

        self.M = M
        self.N = N
        self.K = K
        self.D = D

    def initAlpha(self):
        """ Initialisation of ARD on the weights"""
        alpha = [ s.zeros(self.K,) for m in range(self.M) ]
        for m in range(self.M):
            tmp = bernoulli.rvs(p=0.5, size=self.K)
            tmp[tmp==1] = 1.
            tmp[tmp==0] = 1E5
            alpha[m] = tmp
        return alpha

    def initW_ard(self, alpha=None):
        """ Initialisation of weights in automatic relevance determination prior"""
        if alpha is None:
            alpha = self.initAlpha()
        W = [ s.zeros((self.D[m],self.K)) for m in range(self.M) ]
        for m in range(self.M):
            for k in range(self.K):
                W[m][:,k] = norm.rvs(loc=0, scale=1/s.sqrt(alpha[m][k]), size=self.D[m])
        return W,alpha

    def initW_spikeslab(self, theta, alpha=None):
        """ Initialisation of weights in spike and slab prior"""

        # Simualte ARD precision
        if alpha is None:
            alpha = self.initAlpha()
        else:
            assert not any([0 in a for a in alpha]), 'alpha cannot be zero'

        # Simulate bernoulli variable S
        S = [ s.zeros((self.D[m],self.K)) for m in range(self.M) ]
        for m in range(self.M):

            # Completely vectorised, not sure if it works
            # S[m] = bernoulli.rvs(p=theta[m].flatten(), size=self.D[m]*self.K).reshape((self.D[m],self.K))

            # Partially vectorised
            for k in range(self.K):
                S[m][:,k] = bernoulli.rvs(p=theta[m][:,k], size=self.D[m])
            
            # Unvectorised
            # for d in range(self.D[m]):
            #     for k in range(self.K):
            #         S[m][d,k] = bernoulli.rvs(p=theta[m][d,k], size=1)


        # Simulate gaussian weights W
        W_hat = [ s.empty((self.D[m],self.K)) for m in range(self.M) ]
        W = [ s.empty((self.D[m],self.K)) for m in range(self.M) ]
        for m in range(self.M):
            for k in range(self.K):
                W_hat[m][:,k] = norm.rvs(loc=0, scale=s.sqrt(1./alpha[m][k]), size=self.D[m])
            W[m] = W_hat[m] * S[m]

        return S, W, W_hat, alpha

    def initZ(self):
        """ Initialisation of latent variables"""
        Z = s.empty((self.N,self.K))
        for n in range(self.N):
            for k in range(self.K):
                Z[n,k] = norm.rvs(loc=0, scale=1, size=1)
        return Z

    def initTau(self):
        """ Initialisation of noise precision"""
        return [ uniform.rvs(loc=1,scale=3,size=self.D[m]) for m in range(self.M) ]

    def generateData(self, W, Z, Tau, likelihood, missingness=0.0, missing_view=False):
        """ Initialisation of observations 

        PARAMETERS
        ----------
        W (list of length M where each element is a np array with shape (Dm,K)): weights
        Z (np array with shape (N,K): latent variables
        Tau (list of length M where each element is a np array with shape (Dm,)): precision of the normally-distributed noise
        likelihood (str): type of likelihood
        missingness (float): percentage of missing values
        """

        Y = [ s.zeros((self.N,self.D[m])) for m in range(self.M) ]
        F = [ s.zeros((self.N,self.D[m])) for m in range(self.M) ]

        if likelihood == "gaussian":
            # Vectorised
            for m in range(self.M):
                F[m] = s.dot(Z,W[m].T) + norm.rvs(loc=0, scale=1/s.sqrt(Tau[m]), size=[self.N, self.D[m]])
                Y[m] = F[m]
            # Non-vectorised, slow
            # for m in range(self.M):
                # for n in range(self.N):
                    # for d in range(self.D[m]):
                        # Y[m][n,d] = s.dot(Z[n,:],W[m][d,:].T) + norm.rvs(loc=0,scale=1/s.sqrt(Tau[m][d]))

        # Sample observations using a poisson likelihood
        elif likelihood == "poisson":

            ## Unvectorised
            # for m in range(self.M):
            #     for n in range(self.N):
            #         for d in range(self.D[m]):
            #             f = s.dot(Z[n,:],W[m][d,:].T)
            #             # f = s.dot(Z[n,:],W[m][d,:].T) + norm.rvs(loc=0,scale=s.sqrt(1/Tau[m][d]))
            #             rate = s.log(1+s.exp(f))
            #             # Sample from the Poisson distribution
            #             # Y[m][n,d] = poisson.rvs(rate)
            #             # Use the more likely values
            #             Y[m][n,d] = s.special.round(rate)

            ## Vectorised
            for m in range(self.M):
                F[m] = s.dot(Z,W[m].T)

                # Without noise
                Y[m] = s.special.round(s.log(1.+s.exp(F[m])))

                # With noise, sample from the Poisson distribution
                # Y[m] = poisson.rvs(rate).astype(float)

        # Sample observations using a bernoulli likelihood
        elif likelihood == "bernoulli":
            for m in range(self.M):

                ## Vectorised 
                F[m] = s.dot(Z,W[m].T)

                # without noise
                Y[m] = s.special.round(sigmoid(F[m]))

                # with noise
                # Y[m] = bernoulli.rvs(f).astype(float)

                ## Unvectorised

                # for n in range(self.N):
                    # for d in range(self.D[m]):
                        # Without noise
                        # f = sigmoid( s.dot(Z[n,:],W[m][d,:].T) )

                        # With noise
                        # Y[m][n,d] = bernoulli.rvs(f)
                        # Use the more likely state
                        # Y[m][n,d] = s.special.round(f)


        # Introduce missing values into the data
        if missingness > 0.0:
            for m in range(self.M):
                nas = s.random.choice(range(self.N*self.D[m]), size=int(missingness*self.N*self.D[m]), replace=False)
                tmp = Y[m].flatten()
                tmp[nas] = s.nan
                Y[m] = tmp.reshape((self.N,self.D[m]))
        if missing_view > 0.0:   # percentage of samples missing a view
            # select samples missing one view
            n_missing = s.random.choice(range(self.N), int(missing_view * self.N), replace=False)
            Y[0][n_missing,:] = s.nan

        # Convert data to pandas data frame
        for m in range(self.M):
            Y[m] = pd.DataFrame(data=Y[m])

        # return F,Y
        return Y



from __future__ import division
import numpy.ma as ma
import numpy as np
import warnings
from time import time
import scipy.special as special

# Import manually defined functions
from .variational_nodes import *
from .utils import *
from .nodes import Constant_Node
from .mixed_nodes import Mixed_Theta_Nodes


warnings.filterwarnings('ignore')

"""
Module to define the nodes and the corresponding updates of the model
All nodes belong to the 'Variational_Node' class. They share the following
    methods:
    - precompute: precompute some terms to speed up the calculations
    - calculateELBO: calculate evidence lower bound using current estimates of expectations/params
    - getParameters: return current parameters
    - getExpectations: return current expectations
    - updateParameters: update parameters using current estimates of expectations
    - updateExpectations: update expectations using current estimates of parameters

    attributes:
    - markov_blanket: dictionary that defines the set of nodes that are in the markov blanket of the current node
    - Q: an instance of Distribution() which contains the specification of the variational distribution
    - P: an instance of Distribution() which contains the specification of the prior distribution
    - dim: dimensionality of the node
"""

class Y_Node(Constant_Variational_Node):
    def __init__(self, dim, value):
        Constant_Variational_Node.__init__(self, dim, value)

        # Create a boolean mask of the data to hide missing values
        if type(self.value) != ma.MaskedArray:
            self.mask()

    def precompute(self):
        # Precompute some terms to speed up the calculations
        self.N = self.dim[0] - ma.getmask(self.value).sum(axis=0)
        self.likconst = -0.5*s.sum(self.N)*s.log(2.*s.pi)
        self.means = self.value.mean(axis=0).data

    def mask(self):
        # Mask the observations if they have missing values
        self.value = ma.masked_invalid(self.value)

    def getMask(self):
        return ma.getmask(self.value)

    def calculateELBO(self):
        # Calculate evidence lower bound
        # We use the trick that the update of Tau already contains the Gaussian likelihod.
        # However, it is important that the lower bound is calculated after the update of Tau is performed
        tauQ_param = self.markov_blanket["Tau"].getParameters("Q")
        tauP_param = self.markov_blanket["Tau"].getParameters("P")
        tau_exp = self.markov_blanket["Tau"].getExpectations(expand=False)

        # Important: this assumes that the Tau update has been done beforehand
        lik = self.likconst + 0.5*s.sum(self.N*(tau_exp["lnE"])) - s.dot(tau_exp["E"],tauQ_param["b"]-tauP_param["b"])
        return lik

class Tau_Node(Gamma_Unobserved_Variational_Node):
    def __init__(self, dim, pa, pb, qa, qb, qE=None):
        super(Tau_Node,self).__init__(dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)

    def precompute(self):

        # constant update of Qa
        Y = self.markov_blanket["Y"].getExpectation()
        # mask = ma.getmask(Y)
        mask = self.markov_blanket["Y"].getMask()
        # Y = Y.data

        Qa = self.P.getParameters()['a'] + (Y.shape[0] - mask.sum(axis=0))/2.
        self.Q.params['a'] = Qa

        # constant lower bound term
        self.lbconst = s.sum(self.P.params['a']*s.log(self.P.params['b']) - special.gammaln(self.P.params['a']))

    def updateParameters(self):

        # Collect expectations from other nodes
        Y = self.markov_blanket["Y"].getExpectation()
        # mask = ma.getmask(Y)
        mask = self.markov_blanket["Y"].getMask()
        
        Wtmp = self.markov_blanket["SW"].getExpectations()
        Ztmp = self.markov_blanket["Z"].getExpectations()
        SW, SWW = Wtmp["E"], Wtmp["ESWW"]
        Z, ZZ = Ztmp["E"], Ztmp["E2"]

        # Collect parameters from the P and Q distributions of this node
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb = P['a'], P['b']

        # Mask matrices
        Y = Y.data
        Y[mask] = 0.

        # Calculate temporary terms for the update
        ZW = Z.dot(SW.T)
        ZW[mask] = 0.

        # Calculate terms for the update
        term1 = s.square(Y).sum(axis=0)

        term2 = ZZ.dot(SWW.T)
        term2[mask] = 0
        term2 = term2.sum(axis=0)

        term3 = np.dot(np.square(Z),np.square(SW).T)
        term3[mask] = 0.
        term3 = -term3.sum(axis=0)
        term3 += np.square(ZW).sum(axis=0)

        ZW *= Y  # Warning for developers: ZW becomes ZWY
        term4 = 2.*(ZW.sum(axis=0))

        tmp = term1 + term2 + term3 - term4

        # Perform updates of the Q distribution
        Qb = Pb + tmp/2.

        # Save updated parameters of the Q distribution
        self.Q.setParameters(a=self.Q.params['a'], b=Qb)

    def calculateELBO(self):
        # Collect parameters and expectations from current node
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb, Qa, Qb = P['a'], P['b'], Q['a'], Q['b']
        QE, QlnE = self.Q.expectations['E'], self.Q.expectations['lnE']

        # Do the calculations
        lb_p = self.lbconst + s.sum((Pa-1.)*QlnE) - s.sum(Pb*QE)
        lb_q = s.sum(Qa*s.log(Qb)) + s.sum((Qa-1.)*QlnE) - s.sum(Qb*QE) - s.sum(special.gammaln(Qa))

        return lb_p - lb_q

    def getExpectations(self, expand=True):
        QExp = self.Q.getExpectations()
        if expand:
            N = self.markov_blanket['Z'].dim[0]
            expanded_E = s.repeat(QExp['E'][None, :], N, axis=0)
            expanded_lnE = s.repeat(QExp['lnE'][None, :], N, axis=0)
            return {'E': expanded_E, 'lnE': expanded_lnE}
        else:
            return QExp

    def getExpectation(self, expand=True):
        QExp = self.getExpectations(expand)
        return QExp['E']

class Alpha_Node(Gamma_Unobserved_Variational_Node):
    def __init__(self, dim, pa, pb, qa, qb, qE=None):
        # Gamma_Unobserved_Variational_Node.__init__(self, dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)
        super(Alpha_Node,self).__init__(dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)

    def precompute(self):
        # self.lbconst = self.K * ( self.P.a*s.log(self.P.b) - special.gammaln(self.P.a) )
        # self.lbconst = s.sum( self.P.params['a']*s.log(self.P.params['b']) - special.gammaln(self.P.params['a']) )
        self.factors_axis = 0

    def updateParameters(self):

        # Collect expectations from other nodes
        tmp = self.markov_blanket["SW"].getExpectations()
        ES,EWW = tmp["ES"],tmp["EWW"]

        # Collect parameters from the P and Q distributions of this node
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb = P['a'], P['b']

        # Perform updates
        Qa = Pa + 0.5*ES.shape[0]
        Qb = Pb + 0.5*EWW.sum(axis=0)

        # Save updated parameters of the Q distribution
        self.Q.setParameters(a=Qa, b=Qb)

    def getExpectations(self, expand=False):
        QExp = self.Q.getExpectations()
        QExp['E'] = QExp['E']
        QExp['lnE'] = QExp['lnE']
        if expand:
            D = self.markov_blanket['SW'].dim[0]
            expanded_E = s.repeat(QExp['E'][None, :], D, axis=0)
            expanded_lnE = s.repeat(QExp['lnE'][None, :], D, axis=0)
            return {'E': expanded_E, 'lnE': expanded_lnE}
        else:
            return QExp

    def getExpectation(self, expand=False):
        QExp = self.getExpectations(expand)
        return QExp['E']

    def calculateELBO(self):
        # Collect parameters and expectations
        P,Q = self.P.getParameters(), self.Q.getParameters()
        Pa, Pb, Qa, Qb = P['a'], P['b'], Q['a'], Q['b']
        QE, QlnE = self.Q.getExpectations()['E'], self.Q.getExpectations()['lnE']

        # Do the calculations
        lb_p = (Pa*s.log(Pb)).sum() - special.gammaln(Pa).sum() + ((Pa-1.)*QlnE).sum() - (Pb*QE).sum()
        lb_q = (Qa*s.log(Qb)).sum() - special.gammaln(Qa).sum() + ((Qa-1.)*QlnE).sum() - (Qb*QE).sum()

        return lb_p - lb_q

class SW_Node(BernoulliGaussian_Unobserved_Variational_Node):
    # TOO MANY ARGUMENTS, SHOULD WE USE **KWARGS AND *KARGS ONLY?
    def __init__(self, dim, pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta, qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0=None, qEW_S1=None, qES=None):
        super(SW_Node,self).__init__(dim, pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta, qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0, qEW_S1, qES)

    def precompute(self):
        self.factors_axis = 1

    def updateParameters(self):
        # Collect expectations from other nodes
        Ztmp = self.markov_blanket["Z"].getExpectations()
        Z,ZZ = Ztmp["E"],Ztmp["E2"]
        tau = self.markov_blanket["Tau"].getExpectation()
        Y = self.markov_blanket["Y"].getExpectation()
        alpha = self.markov_blanket["Alpha"].getExpectation(expand=False)
        thetatmp = self.markov_blanket["Theta"].getExpectations()
        theta_lnE, theta_lnEInv  = thetatmp['lnE'], thetatmp['lnEInv']
        mask = self.markov_blanket["Y"].getMask()
        # mask = ma.getmask(Y)

        # Collect parameters and expectations from P and Q distributions of this node
        SW = self.Q.getExpectations()["E"]
        Q = self.Q.getParameters()
        Qmean_S1, Qvar_S1, Qtheta = Q['mean_S1'], Q['var_S1'], Q['theta']

        # Mask matrices
        # Ymean = Y.mean(axis=0)
        Y = Y.data
        Y[mask] = 0.
        tau[mask] = 0.

        # precompute terms used for all factors
        tauYT = (tau*Y).T

        for k in range(self.dim[1]):
            # Calculate intermediate steps
            term1 = (theta_lnE-theta_lnEInv)[k]
            term2 = 0.5*s.log(alpha[k])
            term3 = 0.5*s.log(s.dot(ZZ[:,k], tau) + alpha[k])

            term4_tmp1 = s.dot(tauYT,Z[:,k])

            term4_tmp2_1 = SW[:,s.arange(self.dim[1])!=k].T
            term4_tmp2_2 = (Z[:,k]*Z[:,s.arange(self.dim[1])!=k].T).T
            term4_tmp2 = s.dot(term4_tmp2_2, term4_tmp2_1)
            term4_tmp2 *= tau
            term4_tmp2 = term4_tmp2.sum(axis=0)

            term4_tmp3 = s.dot(ZZ[:,k].T,tau) + alpha[k]

            term4 = 0.5*s.divide(s.square(term4_tmp1-term4_tmp2),term4_tmp3)

            # Update S
            # NOTE there could be some precision issues in S --> loads of 1s in result
            Qtheta[:,k] = 1./(1.+s.exp(-(term1+term2-term3+term4)))

            # Update W
            Qvar_S1[:,k] = 1./term4_tmp3
            Qmean_S1[:,k] = Qvar_S1[:,k]*(term4_tmp1-term4_tmp2)

            # Update Expectations for the next iteration
            SW[:,k] = Qtheta[:,k] * Qmean_S1[:,k]

        # Save updated parameters of the Q distribution
        # self.Q.setParameters(mean_S0=0., var_S0=1./alpha, mean_S1=Qmean_S1, var_S1=Qvar_S1, theta=Qtheta )
        self.Q.setParameters(mean_S0=s.zeros((self.dim[0],self.dim[1])), var_S0=s.repeat(1./alpha[None,:],self.dim[0],0), mean_S1=Qmean_S1, var_S1=Qvar_S1, theta=Qtheta )

    def calculateELBO(self):

        # Collect parameters and expectations
        Qpar,Qexp = self.Q.getParameters(), self.Q.getExpectations()
        S,WW = Qexp["ES"], Qexp["EWW"]
        Qvar = Qpar['var_S1']
        theta = self.markov_blanket['Theta'].getExpectations()
        alpha = self.markov_blanket["Alpha"].getExpectations(expand=False)

        # Calculate ELBO for W
        lb_pw = 0.5*(self.dim[0]*alpha["lnE"].sum() - s.sum(alpha["E"]*WW))
        lb_qw = -0.5*self.dim[1]*self.dim[0] - 0.5*(S*s.log(Qvar) + (1.-S)*s.log(1./alpha["E"])).sum()
        lb_w = lb_pw - lb_qw

        # Calculate ELBO for S
        # TO-DO: CHECK THAT THE BROADCASTING IS CORRECT FOR THETA
        lb_ps = S*theta['lnE'] + (1.-S)*theta['lnEInv']
        lb_qs = S*s.log(S) + (1.-S)*s.log(1.-S)
        lb_ps[s.isnan(lb_ps)] = 0.
        lb_qs[s.isnan(lb_qs)] = 0.
        lb_s = s.sum(lb_ps) - s.sum(lb_qs)

        return lb_w + lb_s

class Theta_Node(Beta_Unobserved_Variational_Node):
    """
    This class comtain a Theta node associate to factors for which
    we dont have annotations.

    The inference is done per view and factor, so the dimension of the node is the
    number of non-annotated factors

    the updateParameters function needs to know what factors are non-annotated in
    order to choose from the S matrix
    """

    def __init__(self, dim, pa, pb, qa, qb, qE=None):
        # Beta_Unobserved_Variational_Node.__init__(self, dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)
        super(Theta_Node,self).__init__(dim=dim, pa=pa, pb=pb, qa=qa, qb=qb, qE=qE)

    def precompute(self):
        self.factors_axis = 0
        self.Ppar = self.P.getParameters()

    def updateParameters(self, factors_selection=None):
        # factors_selection (np array or list): indices of factors that are non-annotated

        # Collect expectations from other nodes
        S = self.markov_blanket['SW'].getExpectations()["ES"]

        # Precompute terms
        if factors_selection is not None:
            tmp1 = S[:,factors_selection].sum(axis=0)
        else:
            tmp1 = S.sum(axis=0)

        # Perform updates
        Qa = self.Ppar['a'] + tmp1
        Qb = self.Ppar['b'] + S.shape[0]-tmp1
        
        # Save updated parameters of the Q distribution
        self.Q.setParameters(a=Qa, b=Qb)

    def calculateELBO(self):
        # TO-DO: IS THIS CORRECT? DO WE NEED TO SUM OVER D?
        
        # Collect parameters and expectations
        Qpar,Qexp = self.getParameters(), self.getExpectations()
        Pa, Pb, Qa, Qb = self.Ppar['a'], self.Ppar['b'], Qpar['a'], Qpar['b']
        QE, QlnE, QlnEInv = Qexp['E'], Qexp['lnE'], Qexp['lnEInv']

        # minus cross entropy of Q and P
        lb_p = (Pa-1.)*QlnE + (Pb-1.)*QlnEInv - special.betaln(Pa,Pb)
        lb_p[np.isnan(lb_p)] = 0

        # minus entropy of Q
        lb_q = (Qa-1.)*QlnE + (Qb-1.)*QlnEInv - special.betaln(Qa,Qb)
        lb_q[np.isnan(lb_q)] = 0

        return lb_p.sum() - lb_q.sum()

class Theta_Constant_Node(Constant_Variational_Node):
    """
    Dimensions of Theta_Constant_Node should be (D[m], K)
    """
    def __init__(self, dim, value):
        super(Theta_Constant_Node, self).__init__(dim, value)

    def precompute(self):
        self.E = self.value
        self.lnE = s.log(self.value)
        self.lnEInv = s.log(1.-self.value)

    def getExpectations(self):
        return { 'E':self.E, 'lnE':self.lnE, 'lnEInv':self.lnEInv }

    def removeFactors(self, idx, axis=1):
        # Ideally we want this node to use the removeFactors defined in Node()
        # but the problem is that we also need to update the "expectations", so i need
        # to call precompute()
        self.value = s.delete(self.value, idx, axis)
        self.precompute()
        self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))

class Z_Node(UnivariateGaussian_Unobserved_Variational_Node):
    def __init__(self, dim, pmean, pvar, qmean, qvar, qE=None, qE2=None, idx_covariates=None):
        super(Z_Node,self).__init__(dim=dim, pmean=pmean, pvar=pvar, qmean=qmean, qvar=qvar, qE=qE, qE2=qE2)

        self.covariates = np.zeros(self.dim[1], dtype=bool)

        # Define indices for covariates
        if idx_covariates is not None:
            self.covariates[idx_covariates] = True

    def precompute(self):
        # Precompute terms to speed up computation
        self.N = self.dim[0]
        self.factors_axis = 1

    def getLvIndex(self):
        # Method to return the index of the latent variables (without covariates)
        latent_variables = np.array(range(self.dim[1]))
        if any(self.covariates):
            # latent_variables = np.delete(latent_variables, latent_variables[self.covariates])
            latent_variables = latent_variables[~self.covariates]
        return latent_variables

    def updateParameters(self):

        # Collect expectations from the markov blanket
        Y = self.markov_blanket["Y"].getExpectation()
        SWtmp = self.markov_blanket["SW"].getExpectations()
        tau = self.markov_blanket["Tau"].getExpectation()
        latent_variables = self.getLvIndex() # excluding covariates from the list of latent variables
        # mask = [ ma.getmask(Y[m]) for m in range(len(Y)) ]
        mask = self.markov_blanket["Y"].getMask()

        # Collect parameters from the prior or expectations from the markov blanket
        Alpha = 1./self.P.getParameters()["var"]

        # Mask data
        for m in range(len(Y)):
            # Mask tau
            tau[m][mask[m]] = 0.
            # Mask Y
            Y[m] = Y[m].data
            Y[m][mask[m]] = 0.

        # Collect parameters from the P and Q distributions of this node
        Q = self.Q.getParameters().copy()
        Qmean, Qvar = Q['mean'], Q['var']

        M = len(Y)
        for k in latent_variables:
            foo = s.zeros((self.N,))
            bar = s.zeros((self.N,))
            for m in range(M):
                foo += np.dot(tau[m], SWtmp[m]["ESWW"][:,k])

                bar_tmp1 = SWtmp[m]["E"][:,k]

                bar_tmp2 = - s.dot(Qmean[:, s.arange(self.dim[1]) != k], SWtmp[m]["E"][:, s.arange(self.dim[1]) != k].T)
                bar_tmp2 += Y[m]
                bar_tmp2 *= tau[m]
                bar += np.dot(bar_tmp2, bar_tmp1)

            Qvar[:,k] = 1./(Alpha[:,k]+foo)
            Qmean[:,k] = Qvar[:,k] * bar

        # Save updated parameters of the Q distribution
        self.Q.setParameters(mean=Qmean, var=Qvar)

    def calculateELBO(self):
        # Collect parameters and expectations of current node
        Qpar,Qexp = self.Q.getParameters(), self.Q.getExpectations()
        Qmean, Qvar = Qpar['mean'], Qpar['var']
        QE, QE2 = Qexp['E'],Qexp['E2']
        Alpha = { 'E':1./self.P.getParameters()["var"], 'lnE':s.log(1./self.P.getParameters()["var"]) }

        # This ELBO term contains only cross entropy between Q and P,and entropy of Q. So the covariates should not intervene at all
        latent_variables = self.getLvIndex()
        Alpha["E"], Alpha["lnE"] = Alpha["E"][:,latent_variables], Alpha["lnE"][:,latent_variables]
        Qmean, Qvar = Qmean[:, latent_variables], Qvar[:, latent_variables]
        QE, QE2 = QE[:, latent_variables], QE2[:, latent_variables]

        # compute term from the exponential in the Gaussian
        QE2 *= 0.5
        tmp1 = -(QE2 * Alpha['E']).sum()

        # compute term from the precision factor in front of the Gaussian
        tmp2 = 0.5*Alpha["lnE"].sum()

        lb_p = tmp1 + tmp2
        lb_q = -0.5*(s.log(Qvar).sum() + self.N*len(latent_variables))

        return lb_p-lb_q




"""
Module to initalise the nodes
"""

import scipy as s
import scipy.stats as stats
from sys import path
import sklearn.decomposition

from .nodes import *
from .multiview_nodes import *
from .nongaussian_nodes import *
from .updates import *


class initModel(object):
    def __init__(self, dim, data, lik, seed=None):
        """
        PARAMETERS
        ----------
         dim: dictionary
            keyworded dimensionalities: N for the number of samples, M for the number of views, K for the number of latent variables, D for the number of features per view (a list)
         data: list of ndarrays of length M: 
            observed data
         lik: list of strings 
            likelihood for each view
        """
        self.data = data
        self.lik = lik
        self.N = dim["N"]
        self.K = dim["K"]
        self.M = dim["M"]
        self.D = dim["D"]

        self.nodes = {}

        # Set the seed
        s.random.seed(seed)

    def initZ(self, pmean, pvar, qmean, qvar, qE=None, qE2=None, covariates=None, scale_covariates=None):
        """Method to initialise the latent variables

        PARAMETERS
        ----------
        pmean:
        pvar:
        qmean
        qvar
        qE
        qE2
        covariates: nd array
            matrix of covariates with dimensions (nsamples,ncovariates)
        scale_covariates: 
        """

        # Initialise mean of the Q distribution
        if qmean is not None:
            if isinstance(qmean,str):
                if qmean == "random": # Random initialisation of latent variables
                    qmean = stats.norm.rvs(loc=0, scale=1, size=(self.N,self.K))

                elif qmean == "orthogonal": # Latent variables are initialised randomly but ensuring orthogonality
                    pca = sklearn.decomposition.PCA(n_components=self.K, copy=True, whiten=True)
                    pca.fit(stats.norm.rvs(loc=0, scale=1, size=(self.N,9999)).T)
                    qmean = pca.components_.T

                elif qmean == "pca": # Latent variables are initialised from PCA in the concatenated matrix
                    pca = sklearn.decomposition.PCA(n_components=self.K, copy=True, whiten=True)
                    pca.fit(s.concatenate(self.data,axis=1).T)
                    qmean = pca.components_.T

            elif isinstance(qmean,s.ndarray):
                assert qmean.shape == (self.N,self.K)

            elif isinstance(qmean,(int,float)):
                qmean = s.ones((self.N,self.K)) * qmean

            else:
                print("Wrong initialisation for Z")
                exit()

        # Add covariates
        if covariates is not None:
            assert scale_covariates != None, "If you use covariates also define data_opts['scale_covariates']"

            # Select indices for covaraites
            idx_covariates = s.array(range(covariates.shape[1]))

            # Center and scale the covariates to match the prior distribution N(0,1)
            # to-do: this needs to be improved to take the particular mean and var into account
            # covariates[scale_covariates] = (covariates - covariates.mean(axis=0)) / covariates.std(axis=0)
            scale_covariates = s.array(scale_covariates)
            covariates[:,scale_covariates] = (covariates[:,scale_covariates] - s.nanmean(covariates[:,scale_covariates], axis=0)) / s.nanstd(covariates[:,scale_covariates], axis=0)

            # Set to zero the missing values in the covariates
            covariates[s.isnan(covariates)] = 0.
            qmean[:,idx_covariates] = covariates
        else:
            idx_covariates = None


        # Initialise the node
        # self.Z = Constant_Node(dim=(self.N,self.K), value=qmean)
        self.Z = Z_Node(dim=(self.N,self.K),
                        pmean=s.ones((self.N,self.K))*pmean,
                        pvar=s.ones((self.K,))*pvar,
                        qmean=s.ones((self.N,self.K))*qmean,
                        qvar=s.ones((self.N,self.K))*qvar,
                        qE=qE, 
                        qE2=qE2,
                        idx_covariates=idx_covariates)
        self.nodes["Z"] = self.Z

    def initSW(self, pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta, qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0, qEW_S1, qES):
        """Method to initialise the spike-slab variable (product of bernoulli and gaussian variables)

        PARAMETERS
        ----------
        """
        SW_list = [None]*self.M
        for m in range(self.M):

           # Initialise first moment
            if isinstance(qmean_S1[m],str):
                if qmean_S1[m] == "random":
                    qmean_S1[m] = stats.norm.rvs(loc=0, scale=1, size=(self.D[m],self.K))
                else:
                    print("%s initialisation not implemented for SW" % qmean_S1[m])
                    exit()
            elif isinstance(qmean_S1[m],s.ndarray):
                assert qmean_S1[m].shape == (self.D[m],self.K), "Wrong dimensionality"
            elif isinstance(qmean_S1[m],(int,float)):
                qmean_S1[m] = s.ones((self.D[m],self.K)) * qmean_S1[m]
            else:
                print("Wrong initialisation for SW")
                exit()

            SW_list[m] = SW_Node(
                dim=(self.D[m],self.K),

                ptheta=ptheta[m],
                pmean_S0=pmean_S0[m],
                pvar_S0=pvar_S0[m],
                pmean_S1=pmean_S1[m],
                pvar_S1=pvar_S1[m],

                qtheta=qtheta[m],
                qmean_S0=qmean_S0[m],
                qvar_S0=qvar_S0[m],
                qmean_S1=qmean_S1[m],
                qvar_S1=qvar_S1[m],

                qES=qES[m],
                qEW_S0=qEW_S0[m],
                qEW_S1=qEW_S1[m],
                )

        self.SW = Multiview_Variational_Node(self.M, *SW_list)
        self.nodes["SW"] = self.SW

    def initAlpha(self, pa, pb, qa, qb, qE):
        """Method to initialise the precision of the group-wise ARD prior

        PARAMETERS
        ----------
         pa: float 
            'a' parameter of the prior distribution
         pb :float
            'b' parameter of the prior distribution
         qb: float
            initialisation of the 'b' parameter of the variational distribution
         qE: float
            initial expectation of the variational distribution
        """
        
        alpha_list = [None]*self.M
        for m in range(self.M):
            alpha_list[m] = Alpha_Node(dim=(self.K,), pa=pa[m], pb=pb[m], qa=qa[m], qb=qb[m], qE=qE[m])
        self.Alpha = Multiview_Variational_Node(self.M, *alpha_list)
        self.nodes["Alpha"] = self.Alpha

    def initTau(self, pa, pb, qa, qb, qE):
        # Method to initialise the precision of the noise
        # Inputs:
        #  pa (float): 'a' parameter of the prior distribution
        #  pb (float): 'b' parameter of the prior distribution
        #  qb (float): initialisation of the 'b' parameter of the variational distribution
        #  qE (float): initial expectation of the variational distribution
        tau_list = [None]*self.M
        for m in range(self.M):
            if self.lik[m] == "poisson":
                tmp = 0.25 + 0.17*s.nanmax(self.data[m],axis=0)
                tau_list[m] = Constant_Node(dim=((self.N,self.D[m])), value=s.repeat(tmp[None,:],self.N,0))

            elif self.lik[m] == "bernoulli":
                # seeger
                # tau_list[m] = Constant_Node(dim=((self.N,self.D[m])), value=0.25)
                # Jaakkola
                tau_list[m] = Tau_Jaakkola(dim=((self.N,self.D[m])), value=1.)
            elif self.lik[m] == "binomial":
                print("Not implemented")
                # tmp = 0.25*s.amax(self.data["tot"][m],axis=0)
                # tau_list[m] = Constant_Node(dim=(self.D[m],), value=tmp)
            elif self.lik[m] == "gaussian":
                tau_list[m] = Tau_Node(dim=(self.D[m],), pa=pa[m], pb=pb[m], qa=qa[m], qb=qb[m], qE=qE[m])
        self.Tau = Multiview_Mixed_Node(self.M,*tau_list)
        self.nodes["Tau"] = self.Tau

    def initY(self):
        # Method to initialise the observed data
        Y_list = [None]*self.M
        for m in range(self.M):
            if self.lik[m]=="gaussian":
                Y_list[m] = Y_Node(dim=(self.N,self.D[m]), value=self.data[m])
            elif self.lik[m]=="poisson":
                # What is the best way to initialise the pseudodata??
                Y_list[m] = Poisson_PseudoY_Seeger(dim=(self.N,self.D[m]), obs=self.data[m], E=None)
            elif self.lik[m]=="bernoulli":
                # Seeger lower bound
                # Y_list[m] = Bernoulli_PseudoY_Seeger(dim=(self.N,self.D[m]), obs=self.data[m], E=None)
                # Jaakkola lower bound
                Y_list[m] =  Bernoulli_PseudoY_Jaakkola(dim=(self.N,self.D[m]), obs=self.data[m], E=None)
        self.Y = Multiview_Mixed_Node(self.M, *Y_list)
        self.nodes["Y"] = self.Y

    def initThetaMixed(self, pa, pb, qa, qb, qE, sparsity):
        # Method to initialie a general theta node
        # Inputs:
        #  pa (float): 'a' parameter of the prior distribution
        #  pb (float): 'b' parameter of the prior distribution
        #  qb (float): initialisation of the 'b' parameter of the variational distribution
        #  qE (float): initial expectation of the variational distribution
        #  sparsity (binary): list with binary matrices with dim (D[m],K)

        Theta_list = [None] * self.M
        for m in range(self.M):
            
            # Initialise constant node
            Kconst = sparsity[m]==0
            if Kconst.sum() == 0:
                ConstThetaNode = None
            else:
                # ConstThetaNode = Theta_Constant_Node(dim=(self.D[m],s.sum(Kconst),), value=qE[m][:,Kconst])
                ConstThetaNode = Theta_Constant_Node(dim=(s.sum(Kconst),), value=qE[m][Kconst])
                Theta_list[m] = ConstThetaNode

            # Initialise non-constant node
            Klearn = sparsity[m]==1
            if Klearn.sum() == 0:
                LearnThetaNode = None
            else:
                # FOR NOW WE JUST TAKE THE FIRST ROW BECAUSE IT IS EXPANDED. IT IS UGLY AS HELL
                LearnThetaNode = Theta_Node(dim=(s.sum(Klearn),), pa=pa[m][Klearn], pb=pb[m][Klearn], qa=qa[m][Klearn], qb=qb[m][Klearn], qE=qE[m][Klearn])
                Theta_list[m] = LearnThetaNode

            # Initialise mixed node
            if (ConstThetaNode is not None) and (LearnThetaNode is not None):
                Theta_list[m] = Mixed_Theta_Nodes(LearnTheta=LearnThetaNode, ConstTheta=ConstThetaNode, idx=sparsity[m])

        self.Theta = Multiview_Mixed_Node(self.M, *Theta_list)
        self.nodes["Theta"] = self.Theta

    def initThetaLearn(self, pa, pb, qa, qb, qE):
        # Method to initialise the theta node
        Theta_list = [None] * self.M
        for m in range(self.M):
            Theta_list[m] = Theta_Node(dim=(self.K,), pa=pa[m], pb=pb[m], qa=qa[m], qb=qb[m], qE=qE[m][0,:])
        self.Theta = Multiview_Variational_Node(self.M, *Theta_list)
        self.nodes["Theta"] = self.Theta

    def initThetaConst(self, value):
        # Method to initialise a constant theta node
        Theta_list = [None] * self.M
        for m in range(self.M):
            Theta_list[m] = Theta_Constant_Node(dim=(self.K,), value=value[m])
        self.Theta = Multiview_Constant_Node(self.M, *Theta_list)
        self.nodes["Theta"] = self.Theta

    def initMuZ(self, clusters=None, pmean=0, pvar=1, qmean=0, qvar=1, qE=None):
        if clusters is None:
            clusters = s.zeros(self.N, int)
        self.MuZ = MuZ_Node_Gaussian(pmean, pvar, qmean, qvar, clusters, self.K)
        # self.Clusters = Constant_Node(pmean, pvar, qmean, qvar, clusters, self.K)
        self.nodes['MuZ'] = self.MuZ

    def initExpectations(self, *nodes):
        # Method to initialise the expectations of some nodes
        for node in nodes:
            self.nodes[node].updateExpectations()

    def getNodes(self):
        return { k:v for (k,v) in self.nodes.items()}



import argparse
import pandas as pd
import scipy as s
import sys
from time import sleep

from .build_model import *
from .utils import *

"""
Entry point for the MOFA model

The order should be:
(1) Set the data (set_data)
(1) Set model options (set_model_options)
(2) Set data options (set_data_options)
(3) Parse the data (parse_data)
(4) Set train options (set_train_options)
(5) define priors and intiialisations (define_priors and define_init)
(6) parse_intercept
(7) Train the model (train_model)
(8) Save the model (save_model)
"""

class entry_point():
  def __init__(self):
    self.print_banner()
    self.dimensionalities = {}

    self.data_opts = {}
    self.model_opts = {}
    self.train_opts = {}

    # Covariates are depreciated
    self.data_opts['covariates'] = None
    self.data_opts['scale_covariates'] = False

  def print_banner(self):
    """ Method to print the MOFA banner """

    banner = """
    ###########################################################
    ###                 __  __  ___  _____ _                ### 
    ###                |  \/  |/ _ \|  ___/ \               ### 
    ###                | |\/| | | | | |_ / _ \              ### 
    ###                | |  | | |_| |  _/ ___ \             ### 
    ###                |_|  |_|\___/|_|/_/   \_\            ### 
    ###                                                     ###
    ########################################################### 
   \n 
    """

    print(banner)
    # sleep(2)

  def set_data(self, data):
    """ Method to set the data 

    PARAMETERS
    ----------
    data: several options:
    - a dictionary where each key is the view names and the object is a numpy array or a pandas data frame
    - a list where each element is a numpy array or a pandas data frame
    """

    # Sanity check
    if isinstance(data, dict):
      data = list(data.values())
    elif isinstance(data, list):
      pass
    else:
      print("Error: Data not recognised")
      sys.stdout.flush()
      exit()

    assert s.all([ isinstance(data[m],s.ndarray) or isinstance(data[m],pd.DataFrame) for m in range(len(data)) ]), "Error, input data is not a numpy.ndarray"

    # Verbose message
    for m in range(len(data)):
      print("Loaded view %d with %d samples and %d features..." % (m, data[m].shape[0], data[m].shape[1]))

    # Doing QC on the data
    self.data = qcData(data)

    # Save dimensionalities
    self.dimensionalities["M"] = len(self.data)
    self.dimensionalities["N"] = self.data[0].shape[0]
    self.dimensionalities["D"] = [self.data[m].shape[1] for m in range(len(self.data))]


  def parse_data(self):
    """ Method to parse the data """

    # Parsing the data: centering, scaling, etc.
    self.parsed_data = parseData(self.data, self.data_opts)

    # Remove samples with missing views
    if self.data_opts['RemoveIncompleteSamples']:
      self.data = removeIncompleteSamples(self.data)
      self.parsed_data = removeIncompleteSamples(self.parsed_data)
      self.dimensionalities["N"] = self.parsed_data[0].shape[0] # Update dimensionalities

  def set_train_options(self, iter=5000, elbofreq=1, startSparsity=100, tolerance=0.01, 
    startDrop=5, freqDrop=1, endDrop=9999, dropR2=0, nostop=False, verbose=False, seed=None
    ):
    """ Set training options """


    # Maximum number of iterations
    self.train_opts['maxiter'] = int(iter)

    # Lower bound computation frequency
    self.train_opts['elbofreq'] = int(elbofreq)

    # Verbosity
    self.train_opts['verbose'] = verbose

    # Criteria to drop latent variables while training
    self.train_opts['drop'] = { "by_r2":float(dropR2), "by_norm":1e-10 }
    self.train_opts['startdrop'] = int(startDrop)
    self.train_opts['freqdrop'] = int(freqDrop)
    self.train_opts['enddrop'] = int(endDrop)
    # print("\nDropping factors with minimum threshold of {0}% variance explained".format(dropR2*100))


    # Tolerance level for convergence
    self.train_opts['tolerance'] = float(tolerance)

    # Do no stop even when convergence criteria is met
    self.train_opts['forceiter'] = nostop

    # Iteration to activate spike and slab sparsity
    self.train_opts['startSparsity'] = int(startSparsity)
    if hasattr(self, 'model_opts'):
      if self.model_opts["sparsity"] is False:  self.train_opts['startSparsity'] = 999999999

    # Define schedule of updates
    self.train_opts['schedule'] = ( "Y", "SW", "Z", "Alpha", "Theta", "Tau" )

    # Seed
    if seed is None:
      seed = 0
    self.train_opts['seed'] = int(seed)

  def set_model_options(self, factors, likelihoods, sparsity=True, learnIntercept=False):
    """ Set model options 

        PARAMETERS
        ----------
        factors: int
          initial number of factors
        likelihoods: character or list of characters
          likelihood per view. Choose from 'gaussian', 'poisson' and 'bernoulli'
        sparsity: bool
          Fraction of values to mask at random
        learnIntercept: bool
          Number of samples to mask at random
    """

    # Define initial number of latent factors
    K = self.dimensionalities["K"] = self.model_opts['factors'] = int(factors)
    M = self.dimensionalities["M"]
    
    # Define likelihoods
    self.model_opts['likelihoods'] = likelihoods
    if type(self.model_opts['likelihoods']) is not list:
      self.model_opts['likelihoods'] = [self.model_opts['likelihoods']]
    assert len(self.model_opts['likelihoods'])==M, "Please specify one likelihood for each view"
    assert set(self.model_opts['likelihoods']).issubset(set(["gaussian","bernoulli","poisson"]))

    # Define whether to learn the feature-wise means
    self.model_opts["learnIntercept"] = learnIntercept
    if learnIntercept: 
      self.model_opts['factors'] += 1
      self.dimensionalities["K"] += 1
      K += 1

    # Define whether to use spike and slab sparsity or not
    if sparsity:
      self.model_opts['sparsity_bool'] = True
      self.model_opts['sparsity'] = [s.ones(K) for m in range(M)]
    else:
      self.model_opts['sparsity_bool'] = False
      print("\nWarning... sparsity is desactivated, we recommend using it\n")
      self.model_opts['sparsity'] = [s.zeros(K) for m in range(M)]
      if hasattr(self, 'train_opts'): self.train_opts['startSparsity'] = 999999999

  def set_data_options(self, view_names=None, center_features=True, scale_features=False, scale_views=False, 
    maskAtRandom=None, maskNSamples=None, RemoveIncompleteSamples=False
    ):

    """ Set data processing options

        PARAMETERS
        ----------
        center_features: bool
          center the features to zero mean?
        scale_features: bool
          scale the features to zero mean?
        scale_views: bool
          scale the features to unit variance?
        maskAtRandom: numeric
          Fraction of values to mask at random
        maskNSamples: bool
          Number of samples to mask at random
        RemoveIncompleteSamples: bool
          Remove samples that are not profiled for all omics
    """
    # Sanity checks
    assert "likelihoods" in self.model_opts, "Likelihoods not found in model options"

    # Sanity checks
    M = self.dimensionalities["M"]
    if view_names is None:
      self.data_opts['view_names'] = ["view_%d" % m for m in range(1,M+1)] 
    else:
      if isinstance(view_names,str): view_names = [ view_names ]
      assert len(view_names)==M, "Length of view names and number of views do not match"
      self.data_opts['view_names'] = view_names

    # Data processing: center features
    if center_features is True:
      self.data_opts['center_features'] = [ True if l=="gaussian" else False for l in self.model_opts["likelihoods"] ]
    else:
      if not self.model_opts["learnIntercept"]: print("\nWarning... you are not centering the data and not learning the intercept. The model is not going to work...\n")
      self.data_opts['center_features'] = [ False for l in self.model_opts["likelihoods"] ]

    # Data processing: scale views
    if scale_views is True:
      print("Warning: you are scaling the Gaussian views to unit variance. As long as the scale of the different views is not massively different, the model does not require this")
      self.data_opts['scale_views'] = [ True if l=="gaussian" else False for l in self.model_opts["likelihoods"] ]
    else:
      self.data_opts['scale_views'] = [ False for l in self.model_opts["likelihoods"] ]

    # Data processing: scale features
    if scale_features:
      print("Warning: you are scaling Gaussian the features to unit variance. This is only recommended if the difference in variances between features are driven by technical and not biological reasons")
      assert data_opts['scale_views'] is False, "Scale either entire views or features, not both"
      self.data_opts['scale_features'] = [ True if l=="gaussian" else False for l in self.model_opts["likelihoods"] ]
    else:
      self.data_opts['scale_features'] = [ False for l in self.model_opts["likelihoods"] ]


    # Data processing: mask values
    if maskAtRandom is not None:
      self.data_opts['maskAtRandom'] = data_opts['maskAtRandom']
      assert len(self.data_opts['maskAtRandom'])==M, "Length of MaskAtRandom and input files does not match"
    else:
      self.data_opts['maskAtRandom'] = [0]*M

    if maskNSamples is not None:
      self.data_opts['maskNSamples'] = data_opts['maskNSamples']
      assert len(self.data_opts['maskNSamples'])==M, "Length of MaskAtRandom and input files does not match"
    else:
      self.data_opts['maskNSamples'] = [0]*M

    # Remove incomplete samples?
    self.data_opts['RemoveIncompleteSamples'] = RemoveIncompleteSamples

  def define_priors(self):
    """ Define priors of the model"""

    N = self.dimensionalities["N"]
    K = self.dimensionalities["K"]
    M = self.dimensionalities["M"]
    D = self.dimensionalities["D"]
    
    # Latent Variables
    self.model_opts["priorZ"] = { 'mean':s.zeros((N,K)) }
    self.model_opts["priorZ"]['var'] = s.ones((K,))*1.

    # Weights
    self.model_opts["priorSW"] = { 'Theta':[s.nan]*M, 'mean_S0':[s.nan]*M, 'var_S0':[s.nan]*M, 'mean_S1':[s.nan]*M, 'var_S1':[s.nan]*M } # Not required
    # self.model_opts["priorAlpha"] = { 'a':[s.ones(K)*1e-14]*M, 'b':[s.ones(K)*1e-14]*M }
    self.model_opts["priorAlpha"] = { 'a':[s.ones(K)*1e-5]*M, 'b':[s.ones(K)*1e-5]*M }

    # Theta
    self.model_opts["priorTheta"] = { 'a':[s.ones(K,) for m in range(M)], 'b':[s.ones(K,) for m in range(M)] }
    for m in range(M):
      for k in range(K):
        if self.model_opts['sparsity'][m][k]==0:
          self.model_opts["priorTheta"]["a"][m][k] = s.nan
          self.model_opts["priorTheta"]["b"][m][k] = s.nan

    # Tau
    # self.model_opts["priorTau"] = { 'a':[s.ones(D[m])*1e-14 for m in range(M)], 'b':[s.ones(D[m])*1e-14 for m in range(M)] }
    self.model_opts["priorTau"] = { 'a':[s.ones(D[m])*1e-5 for m in range(M)], 'b':[s.ones(D[m])*1e-5 for m in range(M)] }

  def define_init(self, initTheta=1.):
    """ Define Initialisations of the model

        PARAMETERS
        ----------
        initTheta flaot
          initialisation for theta. Default is 1. (no sparsity)
    """

    N = self.dimensionalities["N"]
    K = self.dimensionalities["K"]
    M = self.dimensionalities["M"]
    D = self.dimensionalities["D"]

    # Latent variables
    self.model_opts["initZ"] = { 'mean':"random", 'var':s.ones((K,)), 'E':None, 'E2':None }

    # Tau
    self.model_opts["initTau"] = { 'a':[s.nan]*M, 'b':[s.nan]*M, 'E':[s.ones(D[m])*100 for m in range(M)] }

    # ARD of weights
    self.model_opts["initAlpha"] = { 'a':[s.nan]*M, 'b':[s.nan]*M, 'E':[s.ones(K)*1. for m in range(M)] }

    # Theta
    self.model_opts["initTheta"] = { 'a':[s.ones(K,) for m in range(M)], 'b':[s.ones(K,) for m in range(M)], 'E':[s.nan*s.zeros(K,) for m in range(M)] }
    if type(initTheta) is float:
      self.model_opts['initTheta']['E'] = [s.ones(K,)*initTheta for m in range(M)]
    else:
       print("Error: 'initTheta' must be a float")
       exit()

    for m in range(M):
      for k in range(K):
        if self.model_opts['sparsity'][m][k]==0.:
          self.model_opts["initTheta"]["a"][m][k] = s.nan
          self.model_opts["initTheta"]["b"][m][k] = s.nan

    # Weights
    self.model_opts["initSW"] = { 
      'Theta':[ s.repeat(self.model_opts['initTheta']['E'][m][None,:],self.dimensionalities["D"][m],0) for m in range(M)],
      'mean_S0':[s.zeros((D[m],K)) for m in range(M)],
      'var_S0':[s.nan*s.ones((D[m],K)) for m in range(M)],
      'mean_S1':[s.zeros((D[m],K)) for m in range(M)],
      # 'mean_S1':[stats.norm.rvs(loc=0, scale=1, size=(D[m],K)) for m in range(M)],
      'var_S1':[s.ones((D[m],K)) for m in range(M)],
      'ES':[None]*M, 'EW_S0':[None]*M, 'EW_S1':[None]*M # It will be calculated from the parameters
    }

  def parse_covariates(self):
    """ Parse covariates """

    print("Covariates are not functional")
    exit()

    K = self.dimensionalities["K"]
    M = self.dimensionalities["M"]

    # Define idx for covariate factors
    idx = range(self.data_opts['covariates'].shape[1])

    # Ignore mean and variance in the prior distribution of Z
    # self.model_opts["priorZ"]["mean"][idx] = s.nan
    self.model_opts["priorZ"]["var"][idx] = s.nan

    # Ignore variance in the variational distribution of Z
    # The mean has been initialised to the covariate values
    self.model_opts["initZ"]["var"][idx] = 0.

  def parse_intercept(self):
    """ Parse intercept factor """

    # Sanity checks
    # TO-DO: CHECK THAT MODEL_OPTS AND DATA_OPTS ARE PROPERLY DEFINED

    K = self.dimensionalities["K"]
    M = self.dimensionalities["M"]
    N = self.dimensionalities["N"]

    # If we want to learn the intercept, we add a constant covariate of 1s
    if self.model_opts['learnIntercept']:
      if self.data_opts['covariates'] is not None:
        self.data_opts['covariates'] = s.insert(self.data_opts['covariates'], obj=0, values=1., axis=1)
        self.data_opts['scale_covariates'].insert(0,False)
      else:
        self.data_opts['covariates'] = s.ones((N,1))
        self.data_opts['scale_covariates'] = [False]

      # Parse intercept
      # self.model_opts['factors'] += 1
      # self.dimensionalities["K"] += 1

      # Remove sparsity from the Intercept factor
      # TO-DO: CHECK THAT THE MODEL IS ALREADY NOT SPARSE
      # TO-DO: RECHECK THIS, ITS UGLY
      # stop if not self.model_opts["learnIntercept"] == TRUE

      for m in range(M):

        # Weights
        # if self.model_opts["likelihoods"][m]=="gaussian":
        self.model_opts["initSW"]["mean_S1"][m][:,0] = s.nanmean(self.parsed_data[m], axis=0)
        self.model_opts["initSW"]["var_S1"][m][:,0] = 1e-10

        # Theta
        self.model_opts['sparsity'][m][0] = 0.
        self.model_opts["initSW"]["Theta"][m][:,0] = 1.
        self.model_opts["priorTheta"]['a'][m][0] = s.nan
        self.model_opts["priorTheta"]['b'][m][0] = s.nan
        self.model_opts["initTheta"]["a"][m][0] = s.nan
        self.model_opts["initTheta"]["b"][m][0] = s.nan
        self.model_opts["initTheta"]["E"][m][0] = 1.

  def train_model(self):
    """ Train the model """

    # Sanity checks
    # TO-DO: CHECK THAT DATA_OPTS, TRAIN_OPTS AND MODEL_OPTS ARE PROPERLY DEFINED
    assert hasattr(self, 'data'), "Data has to be defined before training the model"

    print("\nLikelihoods are defined as:")
    for a,b in zip(self.data_opts['view_names'], self.model_opts['likelihoods']):
      print("\t%s: %s" % (a,b))

    sys.stdout.flush()
    self.model = runMOFA(self.parsed_data, self.data_opts, self.model_opts, self.train_opts, self.train_opts['seed'])
    sys.stdout.flush()

  def save_model(self, outfile, sample_names=None, feature_names=None):
    """ Save the model """

    # Sanity checks
    assert hasattr(self, 'data'), "Data has to be defined before training the model"
    assert hasattr(self, 'model'), "No trained model found"

    # Create output directory
    if not os.path.isdir(os.path.dirname(outfile)):
        print("Output directory does not exist, creating it...")
        os.makedirs(os.path.dirname(outfile))

    # Verbose messages
    print("Saving model in %s...\n" % outfile)
    if sample_names is None:
      print("Sample names not provided...")
    if feature_names is None:
      print("Feature names not provided...")

    # Save the model
    saveModel(self.model, 
      data=self.data,      # Uncentered data
      outfile=outfile, 
      view_names=self.data_opts['view_names'],
      sample_names=sample_names, 
      feature_names=feature_names, 
      train_opts=self.train_opts, 
      model_opts=self.model_opts
    )

    sys.stdout.flush()



"""
Module to define general nodes in a Bayesian network

All nodes have two main attributes:
- dim: dimensionality of the node
- markov_blanket: the markov blanket of the node

"""

import scipy as s
import numpy as np


class Node(object):
    """General class for a node in a Bayesian network

    PARAMETERS
    ----------
    dim: tuple
        Dimensionality of the node
    """
    def __init__(self, dim):
        self.dim = dim

    def addMarkovBlanket(self, **kwargs):
        """Method to define the Markov blanket of the node"""
        if hasattr(self, 'markov_blanket'):
            for k,v in kwargs.items():
                if k in self.markov_blanket.keys():
                    print("Error: " + str(k) + " is already in the markov blanket of " + str(self))
                    exit()
                else:
                    self.markov_blanket[k] = v
        else:
            self.markov_blanket = kwargs

    def getMarkovBlanket(self):
        """ Method to return the Markov blanket of the node """
        return self.markov_blanket

    def update(self):
        """ General method to update both parameters and expectations of the node """
        self.updateParameters()
        self.updateExpectations()

    def updateExpectations(self):
        """ General method to update the expectations of a node """
        pass

    def getDimensions(self):
        """ Method to return dimensionality of the node """
        return self.dim

    def getExpectations(self):
        """ General method to get all expectations of a node """
        pass

    def getExpectation(self):
        """ General method to get the first moment (expectation) of a node """
        pass

    def updateParameters(self):
        """ General function to update parameters of the node """
        pass

    def getParameters(self):
        """ General function to get the parameters of the node """
        pass

    def updateDim(self, axis, new_dim):
        """ Method to update the dimensionality of a node 
        PARAMETERS
        ----------
        axis:
        new_dim:
        """
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)

    def precompute(self):
        # General method to do precomputations
        pass
        
class Constant_Node(Node):
    """ General class for a constant node in a Bayesian network
    Constant nodes do not have expectations or parameters but just values.
    However, for technical reasons the constant cvalues are defined as expectations
    """
    def __init__(self, dim, value):
        self.dim = dim
        if isinstance(value,(int,float)):
            self.value = value * s.ones(dim)
        else:
            assert value.shape == dim, "dimensionality mismatch"
            self.value = value

    def getValue(self):
        """ Method to return the values of the node """
        return self.value

    def getExpectation(self):
        """ Method to return the first moment of the node, which just points to the values """
        return self.getValue()

    def getExpectations(self):
        """ Method to return the expectations of the node, which just points to the values """
        return { 'E':self.getValue(), 'lnE':s.log(self.getValue()), 'E2':self.getValue()**2 }

    def removeFactors(self, idx, axis=None):
        if hasattr(self,"factors_axis"): axis = self.factors_axis
        if axis is not None:
            self.value = s.delete(self.value, idx, axis)
            self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))




"""
This module is used to define the class containing the entire Bayesian Network,
and the corresponding attributes/methods to train the model, set algorithmic options, calculate lower bound, etc.

To-do:
- default values for schedule and options
"""

from __future__ import division
from time import time
import os
import scipy as s
import pandas as pd
import sys

from .variational_nodes import Variational_Node
from .utils import corr, nans



class BayesNet(object):
    def __init__(self, dim, nodes, options):
        """ Initialisation of a Bayesian network

        PARAMETERS
        ----------
        dim: dict
            keyworded dimensionalities, ex. {'N'=10, 'M'=3, ...}
        nodes: dict
            dictionary with all nodes where the keys are the name of the node and the values are instances the 'Node' class
        options: dict
            training options, such as schedule of updates, maximum number of iterations, training options, etc.
        """

        self.dim = dim
        self.nodes = nodes
        self.options = options

        # Training flag
        self.trained = False

    def removeInactiveFactors(self, by_norm=0, by_r2=0):
        """Method to remove inactive factors

        PARAMETERS
        ----------
        by_norm: float
            threshold to shut down factors based on the norm of the latent variable
            CURRENTLY NOT IMPLEMENTED
        by_r2: float
            threshold to shut down factors based on the coefficient of determination
        """
        drop_dic = {}

        # Shut down based on norm of latent variable vectors
        if by_norm > 0:
            Z = self.nodes["Z"].getExpectation()
            drop_dic["by_norm"] = s.where((Z**2).mean(axis=0) < by_norm)[0]
            if len(drop_dic["by_norm"]) > 0:
                print("\n...A Factor has a norm smaller than {0}, dropping it and recomputing ELBO...\n".format(by_norm))
                drop_dic["by_norm"] = [ s.random.choice(drop_dic["by_norm"]) ]

        # Shut down based on coefficient of determination with respect to the residual variance
        #   Advantages: it takes into account both weights and latent variables, is based on how well the model fits the data
        #   Disadvantages: slow, doesnt work with non-gaussian data
        if by_r2>0:
            Z = self.nodes['Z'].getExpectation()
            Y = self.nodes["Y"].getExpectation()
            W = self.nodes["SW"].getExpectation()
            all_r2 = s.zeros([self.dim['M'], self.dim['K']])
            initial_k = 0
            for m in range(self.dim['M']):

                # Fetch the mask for missing vlaues
                mask = self.nodes["Y"].getNodes()[m].getMask()

                # Fetch the data
                Ym = Y[m].data.copy()

                # If there is an intercept term, regress it out
                if s.all(Z[:,0]==1.):
                    Ym -= W[m][:,0]
                    all_r2[:,0] = 1.
                    initial_k = 1

                # Subtract mean and compute sum of squares (denominator)
                # Ym[mask] = 0.
                # Ym -= s.mean(Ym, axis=0)
                # Ym[mask] = 0.


                # Compute R2
                SS = (Ym**2.).sum()
                for k in range(initial_k,self.dim['K']):
                    Ypred_mk = s.outer(Z[:,k], W[m][:,k])
                    Ypred_mk[mask] = 0.
                    Res = ((Ym - Ypred_mk)**2.).sum()
                    all_r2[m,k] = 1. - Res/SS

            # Select factor to remove. If multiple, then just pick one at random.
            drop_dic["by_r2"] = s.where( (all_r2>by_r2).sum(axis=0) == 0)[0] 
            # drop_dic["by_r2"] = s.where( ((all_r2)>by_r2+1e-16).sum(axis=0) == 0)[0] 
            if len(drop_dic["by_r2"]) > 0: 
                print("\n...A Factor explains less than {0}% of variance, dropping it and recomputing ELBO...\n".format(by_r2*100))
                drop_dic["by_r2"] = [ s.random.choice(drop_dic["by_r2"]) ] # drop one factor at a time

        # Drop the factors
        drop = s.unique(s.concatenate(list(drop_dic.values())))
        if len(drop) > 0:
            for node in self.nodes.keys():
                self.nodes[node].removeFactors(drop)
            self.dim['K'] -= len(drop)

        # TO-DO: THIS ALSO COUNTS COVARIATES
        if self.dim['K']==0:
            print("Shut down all components, no structure found in the data.")
            sys.stdout.flush()
            exit()

        pass

    def iterate(self):
        """Method to start iterating and updating the variables using the VB algorithm"""

        # Define some variables to monitor training
        nodes = list(self.getVariationalNodes().keys())
        elbo = pd.DataFrame(data = nans((self.options['maxiter'], len(nodes)+1 )), columns = nodes+["total"] )
        activeK = nans((self.options['maxiter']))
        
        # Precompute updates
        for n in self.nodes:
            self.nodes[n].precompute()

        # Start training
        for i in range(self.options['maxiter']):
            t = time();

            # Remove inactive latent variables
            if (i>=self.options["startdrop"]) and (i%self.options['freqdrop'])==0 and i<=self.options["enddrop"]:
                self.removeInactiveFactors(**self.options['drop'])
                activeK[i] = self.dim["K"]

            # Update node by node, with E and M step merged
            for node in self.options["schedule"]:
                if node=="Theta" and i<self.options['startSparsity']:
                    continue
                self.nodes[node].update()

            # if i==self.options['startSparsity'] and "Theta" in self.options["schedule"]:
            #     print("\n...Activating sparsity, recomputing ELBO...\n")

            # Calculate Evidence Lower Bound
            if (i+1) % self.options['elbofreq'] == 0 and activeK[i]==activeK[i-1] and i!=self.options['startSparsity']:
                elbo.iloc[i] = self.calculateELBO()

                # Print first iteration
                if i==0:
                    print("Iteration 1: time=%.2f ELBO=%.2f, Factors=%d" % (time()-t,elbo.iloc[i]["total"], (~self.nodes["Z"].covariates).sum() ))
                    if self.options['verbose']:
                        print("".join([ "%s=%.2f  " % (k,v) for k,v in elbo.iloc[i].drop("total").iteritems() ]) + "\n")

                else:
                    # Check convergence using the ELBO
                    delta_elbo = elbo.iloc[i]["total"]-elbo.iloc[i-self.options['elbofreq']]["total"]

                    # Print ELBO monitoring
                    if not s.isnan(delta_elbo):
                        print("Iteration %d: time=%.2f ELBO=%.2f, deltaELBO=%.4f, Factors=%d" % (i+1, time()-t, elbo.iloc[i]["total"], delta_elbo, (~self.nodes["Z"].covariates).sum() ))
                    if self.options['verbose']:
                        print("".join([ "%s=%.2f  " % (k,v) for k,v in elbo.iloc[i].drop("total").iteritems() ]) + "\n")
                    if delta_elbo<0 and i!=self.options['startSparsity'] and self.options['verbose']: print("Warning, lower bound is decreasing..."); print('\a')

                    # Assess convergence
                    if (0 <= delta_elbo < self.options['tolerance']) and (not self.options['forceiter']):
                        activeK = activeK[:(i+1)]
                        elbo = elbo[:(i+1)]
                        print ("Converged!\n")
                        break

            # Do not calculate lower bound
            else:
                print("Iteration %d: time=%.2f, Factors=%d" % (i+1,time()-t,(~self.nodes["Z"].covariates).sum()))

            # Flush (we need this to print when running on the cluster)
            sys.stdout.flush()

        # Finish by collecting the training statistics
        self.train_stats = { 'activeK':activeK, 'elbo':elbo["total"].values, 'elbo_terms':elbo.drop("total",1) }
        self.trained = True

    def getParameters(self, *nodes):
        """Method to collect all parameters of a given set of nodes (all by default)

        PARAMETERS
        ----------
        nodes: iterable
            name of the nodes
        """

        if len(nodes) == 0: nodes = self.nodes.keys()
        params = {}
        for node in nodes:
            tmp = self.nodes[node].getParameters()
            if tmp != None: params[node] = tmp
        return params

    def getExpectations(self, only_first_moments=False, *nodes):
        """Method to collect all expectations of a given set of nodes (all by default)
        
        PARAMETERS
        ----------
        only_first_moments: bool
            get only first moments?
        nodes: list
            name of the nodes
        """

        if len(nodes) == 0: nodes = self.nodes.keys()
        expectations = {}
        for node in nodes:
            if only_first_moments:
                tmp = self.nodes[node].getExpectation()
            else:
                tmp = self.nodes[node].getExpectations()
            expectations[node] = tmp
        return expectations

    def getNodes(self):
        """ Method to return all nodes """
        return self.nodes

    def getVariationalNodes(self):
        """ Method to return all variational nodes """
        return { k:v for k,v in self.nodes.items() if isinstance(v,Variational_Node) }

    def getTrainingStats(self):
        """ Method to return training statistics """
        return self.train_stats

    def getTrainingOpts(self):
        """ Method to return training options """
        return self.options

    def getTrainingData(self):
        """ Method to return training data """
        return self.nodes["Y"].getValues()

    def calculateELBO(self, *nodes):
        """Method to calculate the Evidence Lower Bound of the model

        PARAMETERS
        ----------
        nodes: iterable
            list/tuple with the name of the nodes
        """
        if len(nodes) == 0: nodes = self.getVariationalNodes().keys()
        elbo = pd.Series(s.zeros(len(nodes)+1), index=list(nodes)+["total"])
        for node in nodes:
            elbo[node] = float(self.nodes[node].calculateELBO())
            elbo["total"] += elbo[node]
        return elbo



"""
This module is used to define classes for statistical distributions

Each 'Distribution' class can store an arbitrary number of distributions of the same type, this is specified in the 'dim' argument

A 'Distribution' class has two main types of attributes: parameters and expectations. Both have to be defined when initialising a class.
Note that in some distributions (Gaussian mainly) a parameter is equal to an expectation. However, they are stored as separate
attributes and are not always necessarily equal due to E and M steps in the VB algorithm.

"""

import scipy as s
import numpy.linalg as linalg
import scipy.special as special
import scipy.stats as stats

from .utils import *



# General class for probability distributions
class Distribution(object):
    """ General class for a statistical distribution """
    def __init__(self, dim):
        self.dim = dim

    def density(self):
        """ General method to calculate density """
        pass
    def loglik(self):
        """ General method to calculate log likelihood """
        pass
    def sample(self):
        """ General method to sample from the distribution """
        pass
    def entropy(self):
        """ General method to calculate entropy """
        pass
    def updateExpectations(self):
        """ General method to update expectations """
        pass

    def getParameters(self):
        """ General getter function for parameters """
        return self.params

    def setParameters(self,**params):
        """ General setter function for parameters """
        self.params = params

    def getExpectation(self):
        """ General getter function for expectations """
        return self.expectations['E']

    def getExpectations(self):
        """ General setter function for expectations """
        return self.expectations

    def CheckDimensionalities(self):
        """ General method to do a sanity check on the dimensionalities """
        # p_dim = set(map(s.shape, self.params.values()))
        e_dim = set(map(s.shape, self.expectations.values()))
        # assert len(p_dim) == 1, "Parameters have different dimensionalities"
        assert len(e_dim) == 1, "Expectations have different dimensionalities"
        # assert e_dim == p_dim, "Parameters and Expectations have different dimensionality"

    def removeDimensions(self, axis, idx):
        """ General method to remove undesired dimensions 

        PARAMETERS
        ----------
        axis: int 
            axis from where to remove the elements
        idx: list or numpy array
            indices of the elements to remove
        """
        assert axis <= len(self.dim)
        assert s.all(idx < self.dim[axis])
        for k in self.params.keys(): self.params[k] = s.delete(self.params[k], idx, axis)
        for k in self.expectations.keys(): self.expectations[k] = s.delete(self.expectations[k], idx, axis)
        self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))

    def updateDim(self, axis, new_dim):
        """ Method to update the dimensionality of a particular axis. This method is a bit inefficient but we store dimensionalities with tuples and they cannot be modified 

        PARAMETERS
        ----------
        axis: int
            axis to be updated
        new_dim: int
            updated dimensionality
        """
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)

# Specific classes for probability distributions
class MultivariateGaussian(Distribution):
    """
    Class to define multivariate Gaussian distribution. 
    This class can store N multivate Gaussian Distributions of dimensionality D each.

    Equations:
    p(X|Mu,Sigma) = 1/(2pi)^{D/2} * 1/(|Sigma|^0.5) * exp( -0.5*(X-Mu)^{T} Sigma^{-1} (X-Mu) )
    log p(X|Mu,Sigma) = -D/2*log(2pi) - 0.5log(|Sigma|) - 0.5*(X-Mu)^{T} Sigma^{-1} (X-Mu)
    E[X] = Mu
    E[X^2] = E[X]E[X] + Cov[X] = MuMu + Sigma
    cov[X] = Sigma
    H[X] = 0.5*log(|Sigma|) + D/2*(1+log(2pi))

    Dimensionalities:
    - X: (N,D)
    - Mu: (N,D)
    - Sigma: (N,D,D)
    - E[X]: (N,D)
    - E[X^2]: (N,D,D)
    """
    def __init__(self, dim, mean, cov, E=None):
        Distribution.__init__(self, dim)

        # Check dimensions are correct
        assert len(dim) == 2, "You need to specify two dimensions for 'dim': (number of distributions, dimensionality) "
        assert not (dim[0]==1 and dim[1]==1), "A 1x1 multivariate normal reduces to a Univariate normal "

        ## Initialise the parameters ##

        # Initialise the mean
        # If 'mean' is a scalar, broadcast it to all dimensions
        if isinstance(mean,(int,float)): mean = s.ones( (dim[0],dim[1]) ) * mean
        # If 'mean' has dim (D,) and we have N distributions, broadcast it to all N distributions
        if len(mean.shape)==1 and mean.shape[0]==dim[1]: mean = s.repeat(mean,dim[0],0)
        assert sum(mean.shape) > 2, "The mean has to be a matrix with shape (N,D) "

        # Initialise the covariance
        # If 'cov' is a matrix and not a tensor, broadcast it along the zeroth axis
        if len(cov.shape) == 2: cov = s.repeat(cov[None,:,:],dim[0],0)
        assert (cov.shape[1]==cov.shape[2]) and (sum(cov.shape[1:])>1), "The covariance has to be a tensor with shape (N,D,D)"

        # Check that the dimensionalities of 'mean' and 'cov' match
        assert cov.shape[1] == mean.shape[1] == dim[1], "Error in the dimensionalities"
        assert cov.shape[0] == mean.shape[0] == dim[0], "Error in the dimensionalities"

        self.params = {'mean':mean, 'cov':cov }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':E }


    def updateExpectations(self):
        # Update first and second moments using current parameters
        E = self.params['mean']

        # self.E2 = s.empty( (self.dim[0],self.dim[1],self.dim[1]) )
        # for i in range(self.dim[0]):
        #     self.E2[i,:,:] = s.outer(self.E[i,:],self.E[i,:]) + self.cov[i,:,:]

        E2 = self.params['cov'].copy()
        for i in range(self.dim[0]):
            E2[i,:,:] += s.outer(E[i,:],E[i,:])

        self.expectations = {'E':E, 'E2':E2}

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return s.sum( stats.multivariate_normal.pdf(x, mean=self.params['mean'][n,:], cov=self.params['cov'][n,:,:]) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        l = 0.
        D = self.dim[1]
        for n in range(self.dim[0]):
            qterm = (x[n,:]-self.params['mean'][n,:]).T.dot(linalg.det(self.params['cov'][n,:,:])).dot(x[n,:]-self.params['mean'][n,:])
            l += -0.5*D*s.log(2*s.pi) - 0.5*s.log(linalg.det(self.params['cov'][n,:,:])) -0.5*qterm
        return l
        # return s.sum( s.log(stats.multivariate_normal.pdf(x, mean=self.mean[n,:], cov=self.cov[n,:,:])) )

    def removeDimensions(self, axis, idx):
        # Method to remove undesired dimensions
        # - axis (int): axis from where to remove the elements
        # - idx (numpy array): indices of the elements to remove
        assert axis <= len(self.dim)
        assert s.all(idx < self.dim[axis])
        self.params["mean"] = s.delete(self.params["mean"], axis=1, obj=idx)
        self.params["cov"] = s.delete(self.params["cov"], axis=1, obj=idx)
        self.params["cov"] = s.delete(self.params["cov"], axis=2, obj=idx)
        self.expectations["E"] = s.delete(self.expectations["E"], axis=1, obj=idx)
        self.expectations["E2"] = s.delete(self.expectations["E2"], axis=1, obj=idx)
        self.expectations["E2"] = s.delete(self.expectations["E2"], axis=2, obj=idx)
        self.dim = (self.dim[0],self.dim[1]-len(idx))

    # def entropy(self):
        # CHECK THIs Is CORRECT
        # tmp = sum( [ logdet(self.cov[i,:,:]) for i in range(self.dim[0]) ] )
        # return ( 0.5*(tmp + (self.dim[0]*self.dim[1])*(1+s.log(2*pi)) ).sum() )
class UnivariateGaussian(Distribution):
    """
    Class to define univariate Gaussian distributions

    Equations:
    Class for a univariate Gaussian distributed node
    p(x|mu,sigma^2) = 1/sqrt(2*pi*sigma^2) * exp(-0.5*(x-mu)^2/(sigma^2) )
    log p(x|mu,sigma^2) =
    E[x] = mu
    var[x] = sigma^2
    H[x] = 0.5*log(sigma^2) + 0.5*(1+log(2pi))

    """
    def __init__(self, dim, mean, var, E=None, E2=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        mean = s.ones(dim) * mean
        var = s.ones(dim) * var
        self.params = { 'mean':mean, 'var':var }

        # Initialise expectations 
        self.expectations = {}
        if E is None:
            self.updateExpectations()
        else:
            self.expectations['E'] = s.ones(dim)*E 

        if E2 is not None:
            self.expectations['E2'] = s.ones(dim)*E2 


        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        # Update first and second moments using current parameters
        E = self.params['mean']
        E2 = E**2 + self.params['var']
        self.expectations = { 'E':E, 'E2':E2 }


    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        # print stats.norm.pdf(x, loc=self.mean, scale=s.sqrt(self.var))
        return s.sum( (1/s.sqrt(2*s.pi*self.params['var'])) * s.exp(-0.5*(x-self.params['mean'])**2/self.params['var']) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        # return s.log(stats.norm.pdf(x, loc=self.mean, scale=s.sqrt(self.var)))
        return s.sum( -0.5*s.log(2*s.pi) - 0.5*s.log(self.params['var']) -0.5*(x-self.params['mean'])**2/self.params['var'] )

    def entropy(self):
        return s.sum( 0.5*s.log(self.params['var']) + 0.5*(1+s.log(2*s.pi)) )
class Gamma(Distribution):
    """
    Class to define Gamma distributions

    Equations:
    p(x|a,b) = (1/Gamma(a)) * b^a * x^(a-1) * e^(-b*x)
    log p(x|a,b) = -log(Gamma(a)) + a*log(b) + (a-1)*log(x) - b*x
    E[x] = a/b
    var[x] = a/b^2
    E[ln(x)] = digamma(a) - ln(b)
    H[x] = ln(Gamma(a)) - (a-1)*digamma(a) - ln(b) + a
    """

    def __init__(self, dim, a, b, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        a = s.ones(dim) * a
        b = s.ones(dim) * b
        self.params = { 'a':a, 'b':b }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':s.ones(dim)*E, 'lnE':s.log(s.ones(dim)*E) }

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params['a']/self.params['b']
        lnE = special.digamma(self.params['a']) - s.log(self.params['b'])
        self.expectations = { 'E':E, 'lnE':lnE }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return s.prod( (1/special.gamma(self.params['a'])) * self.params['b']**self.params['a'] * x**(self.params['a']-1) * s.exp(-self.params['b']*x) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return s.sum( -s.log(special.gamma(self.params['a'])) + self.params['a']*s.log(self.params['b']) * (self.params['a']-1)*s.log(x) -self.params['b']*x )
class Poisson(Distribution):
    """
    Class to define Poisson distributions.

    Equations:
    p(x|theta) = theta**x * exp(-theta) * 1/theta!
    log p(x|a,b) = x*theta - theta - log(x!)
    E[x] = theta
    var[x] = theta
    """

    def __init__(self, dim, theta, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        theta = s.ones(dim) * theta
        self.params = { 'theta':theta }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':s.ones(dim)*E }

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params['theta']
        self.expectations = { 'E':E }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        theta = self.params['theta'].flatten()
        x = x.flatten()
        # return s.prod (stats.poisson.pmf(x,theta) )
        return s.prod( s.divide(theta**x * s.exp(-theta),s.misc.factorial(x)) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        theta = self.params['theta'].flatten()
        x = x.flatten()
        # return s.log( s.prod (stats.poisson.pmf(x,theta) ))
        return s.sum( x*s.log(theta) - theta - s.log(s.misc.factorial(x)) )
class Bernoulli(Distribution):
    """
    Class to define Bernoulli distributions

    Equations:

    """
    def __init__(self, dim, theta, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        theta = s.ones(dim) * theta
        self.params = { 'theta':theta }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':s.ones(dim)*E }

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params['theta']
        self.expectations = { 'E':E }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return s.prod( self.params['theta']**x * (1-self.params['theta'])**(1-x) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return s.sum( x*self.params['theta'] + (1-x)*(1-self.params['theta']) )
class BernoulliGaussian(Distribution):
    """
    Class to define a Bernoulli-Gaussian distributions (for more information see Titsias and Gredilla, 2014)

    The best way to characterise a joint Bernoulli-Gaussian distribution P(w,s) is by considering
    its factorisation p(w|s)p(s) where s is a bernoulli distribution and w|s=0 and w|s=1 are normal distributions

    Equations:
    p(w,s) = Normal(w|mean,var) * Bernoulli(s|theta)
    FINISH EQUATIONS

    ROOM FOR IMPROVEMENT: i think the current code is inefficient because you have to keep track of the params
    and expectations in both the factorised distributions and the joint one. I think
    perhaps is better not to define new Bernoulli and UnivariateGaussian but work directly with the joint model

    """
    def __init__(self, dim, mean_S0, mean_S1, var_S0, var_S1, theta, EW_S0=None, EW_S1=None, ES=None):
        Distribution.__init__(self,dim)
        self.S = Bernoulli(dim=dim, theta=theta, E=ES)
        self.W_S0 = UnivariateGaussian(dim=dim, mean=mean_S0, var=var_S0, E=EW_S0)
        self.W_S1 = UnivariateGaussian(dim=dim, mean=mean_S1, var=var_S1, E=EW_S1)

        # Collect parameters
        self.params = { 'mean_S0':mean_S0, 'mean_S1':mean_S1, 'var_S0':var_S0, 'var_S1':var_S1, 'theta':theta }
        
        # Collect expectations
        self.updateExpectations()

    def setParameters(self,**params):
        # Setter function for parameters
        self.S.setParameters(theta=params['theta'])
        self.W_S0.setParameters(mean=params['mean_S0'], var=params['var_S0'])
        self.W_S1.setParameters(mean=params['mean_S1'], var=params['var_S1'])
        self.params = params

    def updateParameters(self):
        # Method to update the parameters of the joint distribution based on its constituent distributions
        self.params = { 'theta':self.S.params["theta"], 
                        'mean_S0':self.W_S0.params["mean"], 'var_S0':self.W_S0.params["var"],
                        'mean_S1':self.W_S1.params["mean"], 'var_S1':self.W_S1.params["var"] }

    def updateExpectations(self):
        # Method to calculate the expectations based on the current estimates for the parameters

        # Update expectations of the constituent distributions
        self.S.updateExpectations()
        self.W_S0.updateExpectations()
        self.W_S1.updateExpectations()

        # Calculate expectations of the joint distribution
        ES = self.S.getExpectation()
        EW = self.W_S1.getExpectation()
        E = ES * EW
        ESWW = ES * (s.square(EW) + self.params["var_S1"])
        # ESWW = self.params["theta"] * (self.params["mean_S1"]**2 + self.params["var_S1"])
        EWW = ES*(s.square(EW)+self.params["var_S1"]) + (1-ES)*self.params["var_S0"]
        # EWW = self.params["theta"]*(self.params["mean_S1"]**2+self.params["var_S1"]) + (1-self.params["theta"])*self.params["var_S0"]

        # Collect expectations
        self.expectations = {'E':E, 'ES':ES, 'EW':EW, 'ESWW':ESWW, 'EWW':EWW }

    def removeDimensions(self, axis, idx):
        # Method to remove undesired dimensions
        # - axis (int): axis from where to remove the elements
        # - idx (numpy array): indices of the elements to remove
        assert axis <= len(self.dim)
        assert s.all(idx < self.dim[axis])
        self.S.removeDimensions(axis,idx)
        self.W_S0.removeDimensions(axis,idx)
        self.W_S1.removeDimensions(axis,idx)
        self.updateParameters()
        self.updateExpectations()

    def updateDim(self, axis, new_dim):
        # Function to update the dimensionality of a particular axis
        self.S.updateDim(axis,new_dim)
        self.W_S0.updateDim(axis,new_dim)
        self.W_S1.updateDim(axis,new_dim)
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)
class Binomial(Distribution):
    """
    Class to define Binomial distributions

    Equations:
    p(x|N,theta) = binom(N,x) * theta**(x) * (1-theta)**(N-x)
    log p(x|N,theta) = log(binom(N,x)) + x*theta + (N-x)*(1-theta)
    E[x] = N*theta
    var[x] = N*theta*(1-theta)
    """
    def __init__(self, dim, N, theta, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        theta = s.ones(dim)*theta
        N = s.ones(dim)*N
        self.params = { 'theta':theta, 'N':N }

        # Initialise expectations 
        if E is None:
            self.updateExpectations()
        else:
            E = s.ones(dim)*E
            self.expectations = { 'E':E }

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params["N"] * self.params["N"]
        self.expectations = { 'E':E }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        # return s.prod( stats.binom.pmf(x, self.params["N"], self.theta) )
        return s.prod( special.binom(self.params["N"],x) * self.params["theta"]**x * (1-self.params["theta"])**(self.params["N"]-x) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        assert x.dtype == int, "x has to be an integer array"
        # print s.sum (stats.binom.logpmf(x, self.params["N"], self.theta) )
        return s.sum( s.log(special.binom(self.params["N"],x)) + x*s.log(self.params["theta"]) + (self.params["N"]-x)*s.log(1-self.params["theta"]) )
class Beta(Distribution):
    """
    Class to define Beta distributions

    Equations:
    p(x|a,b) = GammaF(a+b)/(GammaF(a)*GammaF(b)) * x**(a-1) * (1-x)**(b-1)
    log p(x|a,b) = log[GammaF(a+b)/(GammaF(a)*GammaF(b))] + (a-1)*x + (b-1)*(1-x)
    E[x] = a/(a+b)
    var[x] = a*b / ((a+b)**2 * (a+b+1))
    """
    def __init__(self, dim, a, b, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        a = s.ones(dim)*a
        b = s.ones(dim)*b
        self.params = { 'a':a, 'b':b }

        # Initialise expectations
        if E is None: 
            self.updateExpectations()
        else:
            self.expectations = { 
            'E':s.ones(dim)*E,
            'lnE':s.log(s.ones(dim)*E),
            'lnEInv':s.log(1.-s.ones(dim)*E)
            }
            self.expectations["lnEInv"][s.isinf(self.expectations["lnEInv"])] = -s.inf

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        a, b = self.params['a'], self.params['b']
        E = s.divide(a,a+b)
        lnE = special.digamma(a) - special.digamma(a+b)
        lnEInv = special.digamma(b) - special.digamma(a+b) # expectation of ln(1-X)
        lnEInv[s.isinf(lnEInv)] = -s.inf # there is a numerical error in lnEInv if E=1
        self.expectations = { 'E':E, 'lnE':lnE, 'lnEInv':lnEInv }

# if __name__ == "__main__":
#     a = Beta(dim=(10,20), a=1, b=1, E=3)
#     MultivariateGaussian(dim=(1,10), mean=stats.norm.rvs(loc=0, scale=1, size=(10,)), cov=s.eye(10,10), E=None)
#     exit()




"""
Module with functions to build the model
"""

import scipy as s
from sys import path
from time import time,sleep
import pandas as pd
import numpy as np
#from joblib import Parallel, delayed

from .init_nodes import *
from .BayesNet import BayesNet
from .utils import *

def runMOFA(data, data_opts, model_opts, train_opts, seed=None):
    """Method to run a MOFA model


    PARAMETERS
    ----------
    data: list of numpy arrays 
    data_opts: dictionary
        data options
    model_opts: dictionary 
        model options
    train_opts: dictionary
        train options
    seed: numeric 
        seed for the random number generator. 
        If None (default), it is generated randomly
    """


    # set the seed
    if seed is None or seed==0:
        seed = int(round(time()*1000)%1e6)
    s.random.seed(seed)



    ####################
    ## Parse the data ##
    ####################

    # Mask
    if 'maskAtRandom' in data_opts or 'maskNSamples' in data_opts:
        if any(data_opts['maskAtRandom']) or any(data_opts['maskNSamples']):
            data = maskData(data, data_opts)

    ######################
    ## Define the model ##
    ######################

    print ("\n")
    print ("#"*24)
    print ("## Building the model ##")
    print ("#"*24)
    print ("\n")
    sleep(1)

    print("Setting random seed %i..." % seed)

    # Define dimensionalities
    M = len(data)
    N = data[0].shape[0]
    D = s.asarray([ data[m].shape[1] for m in range(M) ])
    K = model_opts["factors"]
    dim = {'M':M, 'N':N, 'D':D, 'K':K }

    ## Define and initialise the nodes ##

    init = initModel(dim, data, model_opts["likelihoods"], seed=seed)

    # Observed data
    init.initY()
    
    # Latent variables
    init.initZ(pmean=model_opts["priorZ"]["mean"], pvar=model_opts["priorZ"]["var"],
               qmean=model_opts["initZ"]["mean"], qvar=model_opts["initZ"]["var"], qE=model_opts["initZ"]["E"], qE2=model_opts["initZ"]["E2"],
               covariates=data_opts['covariates'], scale_covariates=data_opts['scale_covariates'])

    # Sparse weights
    init.initSW(ptheta=model_opts["priorSW"]["Theta"], pmean_S0=model_opts["priorSW"]["mean_S0"], pvar_S0=model_opts["priorSW"]["var_S0"], pmean_S1=model_opts["priorSW"]["mean_S1"], pvar_S1=model_opts["priorSW"]["var_S1"],
                qtheta=model_opts["initSW"]["Theta"], qmean_S0=model_opts["initSW"]["mean_S0"], qvar_S0=model_opts["initSW"]["var_S0"], qmean_S1=model_opts["initSW"]["mean_S1"], qvar_S1=model_opts["initSW"]["var_S1"],
                qEW_S0=model_opts["initSW"]["EW_S0"], qEW_S1=model_opts["initSW"]["EW_S1"], qES=model_opts["initSW"]["ES"])

    # ARD on weights
    init.initAlpha(pa=model_opts["priorAlpha"]['a'], pb=model_opts["priorAlpha"]['b'],
                       qa=model_opts["initAlpha"]['a'], qb=model_opts["initAlpha"]['b'], qE=model_opts["initAlpha"]['E'])

    # Precision of noise
    init.initTau(pa=model_opts["priorTau"]['a'], pb=model_opts["priorTau"]['b'],
                 qa=model_opts["initTau"]['a'], qb=model_opts["initTau"]['b'], qE=model_opts["initTau"]['E'])

    # Sparsity on the weights
    if len(s.unique(model_opts['sparsity'])) == 1:

        # All are infered
        if s.unique(model_opts['sparsity'])==1.:
            # init.initThetaLearn(pa=model_opts["priorTheta"]['a'], pb=model_opts["priorTheta"]['b'],
            #     qa=model_opts["initTheta"]['a'],  qb=model_opts["initTheta"]['b'], qE=model_opts["initTheta"]['E'])
            init.initThetaMixed(pa=model_opts["priorTheta"]['a'], pb=model_opts["priorTheta"]['b'],
                qa=model_opts["initTheta"]['a'],  qb=model_opts["initTheta"]['b'], qE=model_opts["initTheta"]['E'],
                sparsity=model_opts['sparsity'])

        # None are infered
        elif s.unique(model_opts['sparsity'])==0.:
            init.initThetaConst(value=model_opts["initTheta"]['E'])

    # Some are infered
    else:
        init.initThetaMixed(pa=model_opts["priorTheta"]['a'], pb=model_opts["priorTheta"]['b'],
            qa=model_opts["initTheta"]['a'],  qb=model_opts["initTheta"]['b'], qE=model_opts["initTheta"]['E'],
            sparsity=model_opts['sparsity'])

    # Define the markov blanket of each node
    nodes = init.getNodes()
    nodes["Z"].addMarkovBlanket(SW=nodes["SW"], Tau=nodes["Tau"], Y=nodes["Y"])
    nodes["Theta"].addMarkovBlanket(SW=nodes["SW"])
    nodes["Alpha"].addMarkovBlanket(SW=nodes["SW"])
    nodes["SW"].addMarkovBlanket(Z=nodes["Z"], Tau=nodes["Tau"], Alpha=nodes["Alpha"], Y=nodes["Y"], Theta=nodes["Theta"])
    nodes["Y"].addMarkovBlanket(Z=nodes["Z"], SW=nodes["SW"], Tau=nodes["Tau"])
    nodes["Tau"].addMarkovBlanket(Z=nodes["Z"], SW=nodes["SW"], Y=nodes["Y"])

    #################################
    ## Create the Bayesian Network ##
    #################################

    # Initialise Bayesian Network
    net = BayesNet(dim=dim, nodes=init.getNodes(), options=train_opts)

    ####################
    ## Start training ##
    ####################

    print ("\n")
    print ("#"*34)
    print ("## Running MOFA with seed %d ##" % seed)
    print ("#"*34)
    print ("\n")
    
    net.iterate()

    print("\n")
    print("#"*43)
    print("## Training finished, processing results ##")
    print("#"*43)
    print("\n")

    return net


from setuptools import setup
from setuptools import find_packages

import sys

def setup_package():
  install_requires = ['pandas', 'scipy', 'numpy', 'sklearn', 'argparse', 'h5py']
  metadata = dict(
      name = 'mofapy',
      version = '1.2',
      description = 'Multi-Omics Factor Analysis, a statistical framework for the integration of multi-omics data',
      url = 'https://github.com/PMBio/MOFA',
      author = 'Ricard Argelaguet, Damien Arnol and Britta Velten',
      author_email = 'ricard.argelaguet@gmail.com',
      license = 'LGPL-3.0',
      packages = find_packages(),
      install_requires = install_requires
    )

  setup(**metadata)

if __name__ == '__main__':
  if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')
    
  setup_package()


"""
Module to define updates for non-conjugate matrix factorisation models.
Reference: 'Fast Variational Bayesian Inference for Non-Conjugate Matrix Factorisation models' by Seeger and Bouchard (2012)

PseudoY: general class for pseudodata
    PseudoY_seeger: general class for pseudodata using seeger aproach
        Poisson_PseudoY: Poisson likelihood
        Bernoulli_PseudoY: Bernoulli likelihood
        Binomial_PseudoY: Binomial likelihood (TO FINISH)
    
    PseudoY_Jaakkola: NOT IMPLEMENTED
        Bernoulli_PseudoY_Jaakkola: bernoulli likelihood for Jaakkola approach (see REF)
        Tau_Jaakkola:

    Warped_PseudoY_Node (TO FINISH)


"""

from __future__ import division
import scipy as s
import numpy.ma as ma

from .variational_nodes import Unobserved_Variational_Node
from .nodes import Node


def sigmoid(X):
    return s.divide(1.,1.+s.exp(-X))
    # return 1./(1.+s.exp(-X))

def lambdafn(X):
    return s.tanh(X/2.)/(4.*X)

##############################
## General pseudodata nodes ##
##############################

class PseudoY(Unobserved_Variational_Node):
    """ General class for pseudodata nodes """
    def __init__(self, dim, obs, params=None, E=None):
        """
        PARAMETERS
        ----------
         dim (2d tuple): dimensionality of each view
         obs (ndarray): observed data
         params:
         E (ndarray): initial expected value of pseudodata
        """
        Unobserved_Variational_Node.__init__(self, dim)

        # Initialise observed data
        assert obs.shape == dim, "Problems with the dimensionalities"
        self.obs = obs

        # Initialise parameters
        if params is not None:
            assert type(self.params) == dict
            self.params = params
        else:
            self.params = {}

        # Create a boolean mask of the data to hidden missing values
        if type(self.obs) != ma.MaskedArray:
            self.mask()

        # Precompute some terms
        # self.precompute()

        # Initialise expectation
        if E is not None:
            assert E.shape == dim, "Problems with the dimensionalities"
            E = ma.masked_invalid(E)
        # else:
            # E = ma.masked_invalid(s.zeros(self.dim))
        self.E = E

    def updateParameters(self):
        pass

    def mask(self):
        # Mask the observations if they have missing values
        self.obs = ma.masked_invalid(self.obs)

    def getMask(self):
        return ma.getmask(self.obs)

    def updateExpectations(self):
        print("Error: expectation updates for pseudodata node depend on the type of likelihood. They have to be specified in a suclass.")
        exit()

    def getExpectation(self):
        return self.E

    def getExpectations(self):
        return { 'E':self.getExpectation() }

    def getObservations(self):
        return self.obs

    def getValue(self):
        return self.obs

    def getParameters(self):
        return self.params

    def calculateELBO(self):
        print("Not implemented")
        exit()

##################
## Seeger nodes ##
##################

class PseudoY_Seeger(PseudoY):
    """ General class for pseudodata nodes using the seeger approach """
    def __init__(self, dim, obs, params=None, E=None):
        # Inputs:
        #  dim (2d tuple): dimensionality of each view
        #  obs (ndarray): observed data
        #  E (ndarray): initial expected value of pseudodata
        PseudoY.__init__(self, dim=dim, obs=obs, params=params, E=E)

    def updateParameters(self):
        Z = self.markov_blanket["Z"].getExpectation()
        SW = self.markov_blanket["SW"].getExpectation()
        self.params["zeta"] = s.dot(Z,SW.T)

    def calculateELBO(self):
        # Compute Lower Bound using the Gaussian likelihood with pseudodata
        # TO-DO: MASK MISSING VALUES???
        Z = self.markov_blanket["Z"].getExpectation()
        SW = self.markov_blanket["SW"].getExpectation()
        tau = self.markov_blanket["Tau"].getExpectation()
        N = Z.shape[0]
        lb = 0.5*(N*ma.sum(s.log(tau)) - ma.sum(tau*(self.E-s.dot(Z,SW.T))**2 )) # (1) tau is of shape (D,) (2) missing a constant term

        # tau_expanded = s.repeat(tau[None,:],N,0)
        # tau_expanded = ma.masked_where(ma.getmask(self.obs), tau_expanded)
        # lb = 0.5*( ma.sum(s.log(tau_expanded)) - ma.sum(tau_expanded*(self.E-s.dot(Z,SW.T))**2 ) ) # (1) tau is of shape (D,) (2) missing a constant term
        return lb

class Poisson_PseudoY_Seeger(PseudoY_Seeger):
    """
    Class for a Poisson pseudodata node.
    Likelihood:
        p(y|x) \prop gamma(x) * e^{-gamma(x)}  (1)
    where gamma(x) is a rate function that is chosen to be convex and log-concave
    A simple choise for the rate function is e^{x} but this rate function is non-robust
    in the presence of outliers, so in Seeger et al they chose the function:
        gamma(x) = log(1+e^x)

    The data follows a Poisson distribution, but Followung Seeger et al the pseudodata Yhat_ij
    follows a normal distribution with mean E[W_{i,:}]*E[Z_{j,:}] and precision 'tau_j'
    where 'tau_j' is an upper bound of the second derivative of the loglikelihood:
        x_ij = sum_k^k w_{i,k}*z_{k,j}
        f_ij''(x_ij) <= tau_j for all i,j

    For the Poisson likelihood with rate function (1), the upper bound tau is calculated as follows:
        f_ij''(x_ij) = 0.25 + 0.17*ymax_j   where ymax_j = max(Y_{:,j})

    Pseudodata is updated as follows:
        yhat_ij = zeta_ij - f'(zeta_ij)/tau_j = ...
    The bound degrades with the presence of entries with large y_ij, so one should consider
    clipping overly large counts

    """
    def __init__(self, dim, obs, params=None, E=None):
        # - dim (2d tuple): dimensionality of each view
        # - obs (ndarray): observed data
        # - E (ndarray): initial expected value of pseudodata
        PseudoY_Seeger.__init__(self, dim=dim, obs=obs, params=params, E=E)

        # Initialise the observed data
        assert s.all(s.mod(self.obs, 1) == 0), "Data must not contain float numbers, only integers"
        assert s.all(self.obs >= 0), "Data must not contain negative numbers"

    def ratefn(self, X):
        # Poisson rate function proposed in Seeger et al.
        return s.log(1.+s.exp(X))

    def clip(self, threshold):
        # The local bound degrades with the presence of large values in the observed data, which should be clipped
        pass

    def updateExpectations(self):
        # Update the pseudodata
        tau = self.markov_blanket["Tau"].getValue()
        self.E = self.params["zeta"] - sigmoid(self.params["zeta"])*(1.-self.obs/self.ratefn(self.params["zeta"])) / tau
        self.means = self.E.mean(axis=0).data
        self.E -= self.means
        # mask = self.getMask()
        # self.E[mask] = s.nan

    def calculateELBO(self):
        """ Compute Lower Bound """

        Wtmp = self.markov_blanket["SW"].getExpectations()
        Ztmp = self.markov_blanket["Z"].getExpectations()
        W, WW = Wtmp["E"], Wtmp["ESWW"]
        Z, ZZ = Ztmp["E"], Ztmp["E2"]
        zeta = self.params["zeta"]
        tau = self.markov_blanket["Tau"].getValue() # to-do: not expand
        mask = self.getMask()

        # Precompute terms
        ZW = Z.dot(W.T)
        ZZWW = s.square(ZW) - s.dot(s.square(Z),s.square(W).T) + ZZ.dot(WW.T)

        # term1 = 0.5*tau*(ZW - zeta)**2
        term1 = 0.5*tau*(ZZWW - 2*ZW*zeta + s.square(zeta))
        term2 = (ZW - zeta)*(sigmoid(zeta)*(1-self.obs/self.ratefn(zeta)))
        term3 = self.ratefn(zeta) - self.obs*s.log(self.ratefn(zeta))

        elbo = -(term1 + term2 + term3)
        elbo[mask] = 0.

        return elbo.sum()

class Bernoulli_PseudoY_Seeger(PseudoY_Seeger):
    """
    Class for a Bernoulli pseudodata node used to model binary data
    Likelihood:
        p(y|x) = (e^{yx}) / (1+e^x)
    Log likelihood:
        f(x) = -log p(y|x) = log(1+e^x) - yx

    The second derivative is upper bounded by tau=0.25

    Folloiwng Seeger et al, the data follows a Bernoulli distribution but the pseudodata follows a
    normal distribution with mean E[W]*E[Z] and precision 'tau'

    TO-DO: IMPROVE EXPLANATION

    Pseudodata is updated as follows:
        yhat_ij = zeta_ij - f'(zeta_ij)/tau
                = zeta_ij - 4*(sigmoid(zeta_ij) - y_ij)

    """
    def __init__(self, dim, obs, params=None, E=None):
        # - dim (2d tuple): dimensionality of each view
        # - obs (ndarray): observed data
        # - params (ndarray): parameters
        # - E (ndarray): initial expected value of pseudodata
        PseudoY_Seeger.__init__(self, dim=dim, obs=obs, params=params, E=E)

        # Initialise the observed data
        assert s.all( (self.obs==0) | (self.obs==1) ), "Data modelled using bernoulli likelihood must be binary, encoded as 0s or 1s"

    def updateExpectations(self):
        # Update the pseudodata
        self.E = self.params["zeta"] - 4.*(sigmoid(self.params["zeta"]) - self.obs)
        self.means = self.E.mean(axis=0).data
        self.E -= self.means
        
    def calculateELBO(self):
        Z = self.markov_blanket["Z"].getExpectation()
        W = self.markov_blanket["SW"].getExpectation()
        mask = self.getMask()
        tmp = s.dot(Z,W.T)
        
        # Compute Lower Bound using the Bernoulli likelihood and the observed data
        lb = self.obs.data*tmp - s.log(1.+s.exp(tmp))
        lb[mask] = 0.

        # Compute Lower Bound using the gaussian likelihood with pseudo data
        # MISSING CONSTANT TERM
        # term1 = 0.5*s.log(self.params["zeta"])
        # term2 = 0.5*self.params["zeta"]*(self.E-tmp)**2
        # lb = term1 - term2
        # lb[mask] = 0.

        return lb.sum()

class Binomial_PseudoY_Seeger(PseudoY_Seeger):
    """
    Class for a Binomial pseudodata node 
    Likelihood:
        p(x|N,theta) = p(x|N,theta) = binom(N,x) * theta**(x) * (1-theta)**(N-x)
        f(x) = -log p(x|N,theta) = -log(binom(N,x)) - x*theta - (N-x)*(1-theta)

    The second derivative is upper bounded:
        f''(x_ij) <= 0.25*max(N_{:,j})

    Folloiwng Seeger et al, the pseudodata yhat_{nd} follows a normal distribution with mean
    E[w_{i,:}]*E[z_{j,:}] and precision 'tau_j'

    IMPROVE EXPLANATION

    Pseudodata is updated as follows
        yhat_ij = zeta_ij - f'(zeta_ij)/tau_j
                = zeta_ij - (N_{ij}*sigmoid(zeta_ij) - y_ij)/tau_d

    """
    def __init__(self, dim, obs, tot, Zeta=None, E=None):
        # - dim (2d tuple): dimensionality of each view
        # - obs (ndarray): observed data
        # - E (ndarray): initial expected value of pseudodata
        PseudoY_Seeger.__init__(self, dim=dim, obs=None, params=params, E=E)

        # Initialise the observed data
        assert s.all(s.mod(obs, 1) == 0) and s.all(s.mod(tot, 1) == 0), "Data must not contain float numbers, only integers"
        assert s.all(obs >= 0) and s.all(tot >= 0), "Data must not contain negative numbers"
        assert s.all(obs <= tot), "Observed counts have to be equal or smaller than the total counts"
        self.obs = obs
        self.tot = tot


    def updateExpectations(self):
        # Update the pseudodata
        tau = self.markov_blanket["Tau"].getValue()
        self.E = self.params["zeta"] - s.divide(self.tot*sigmoid(self.params["zeta"])-self.obs, tau)
        pass

    def calculateELBO(self):
        # Compute Lower Bound using the Bernoulli likelihood with observed data
        Z = self.markov_blanket["Z"].getExpectation()
        SW = self.markov_blanket["SW"].getExpectation()

        tmp = sigmoid(s.dot(Z,SW.T))

        # TODO change apprximation
        tmp[tmp==0] = 0.00000001
        tmp[tmp==1] = 0.99999999
        lik = s.log(s.special.binom(self.tot,self.obs)).sum() + s.sum(self.obs*s.log(tmp)) + \
            s.sum((self.tot-self.obs)*s.log(1-tmp))
        return lik


####################
## Jaakkola nodes ##
####################

class Tau_Jaakkola(Node):
    """
    Local Parameter that needs to be optimised in the Jaakkola approach.
    For more details see Supplementary Methods 
    """
    def __init__(self, dim, value):
        Node.__init__(self, dim=dim)

        if isinstance(value,(int,float)):
            self.value = value * s.ones(dim)
        else:
            assert value.shape == dim, "Dimensionality mismatch"
            self.value = value

    def updateExpectations(self):
        self.value = 2*lambdafn(self.markov_blanket["Y"].getParameters()["zeta"])

    def getValue(self):
        return self.value

    def getExpectation(self):
        return self.getValue()

    def getExpectations(self):
        return { 'E':self.getValue(), 'lnE':s.log(self.getValue()) }

    def removeFactors(self, idx, axis=None):
        pass
class Bernoulli_PseudoY_Jaakkola(PseudoY):
    """
    Class for a Bernoulli pseudodata node using the Jaakkola approach:
    Likelihood:
        p(y|x) = (e^{yx}) / (1+e^x)
    Following Jaakola et al and intterpreting the bound as a likelihood on gaussian pseudodata
    leads to the folllowing updates

    Pseudodata is given by
            yhat_ij = (2*y_ij-1)/(4*lambadfn(xi_ij))
        where lambdafn(x)= tanh(x/2)/(4*x).

    Its conditional distribution is given by
            N((ZW)_ij, 1/(2 lambadfn(xi_ij)))

    Updates for the variational parameter xi_ij are given by
            sqrt(E((ZW)_ij^2))

    xi_ij in above notation is the same as zeta (variational parameter)

    NOTE: For this class to work the noise variance tau needs to be updated according to
        tau_ij <- 2*lambadfn(xi_ij)
    """
    def __init__(self, dim, obs, params=None, E=None):
        PseudoY.__init__(self, dim=dim, obs=obs, params=params, E=E)
        # Initialise the observed data
        assert s.all( (self.obs==0) | (self.obs==1) ), "Data modelled using bernoulli likelihood must be binary, encoded as 0s or 1s"

    def updateExpectations(self):
        self.E = (2.*self.obs - 1.) / (4.*lambdafn(self.params["zeta"]))
        self.means = self.E.mean(axis=0).data
        self.E -= self.means


    def updateParameters(self):
        Z = self.markov_blanket["Z"].getExpectations()
        SW = self.markov_blanket["SW"].getExpectations()
        self.params["zeta"] = s.sqrt( s.square(Z["E"].dot(SW["E"].T)) - s.dot(s.square(Z["E"]),s.square(SW["E"].T)) + s.dot(Z["E2"], SW["ESWW"].T) )

    def calculateELBO(self):
        Z = self.markov_blanket["Z"].getExpectation()
        Wtmp = self.markov_blanket["SW"].getExpectations()
        Ztmp = self.markov_blanket["Z"].getExpectations()
        zeta = self.params["zeta"]
        SW, SWW = Wtmp["E"], Wtmp["ESWW"]
        Z, ZZ = Ztmp["E"], Ztmp["E2"]
        mask = self.getMask()

        # Compute Lower Bound using the Bernoulli likelihood and the observed data
        # BOTH ARE WRONG AS THEY EXCHANGE LOG AND EXPECTATIONS
        # lb = self.obs.data*tmp - s.log(1.+s.exp(tmp))
        # lb = s.log(1.+s.exp(-(2.*self.obs-1)*tmp)) # DAMIEN'S suggestion
        # lb[mask] = 0.

        # Compute Lower Bound using the gaussian likelihood with pseudo data
        # MISSING CONSTANT TERM
        # term1 = 0.5*s.log(self.params["zeta"])
        # term2 = 0.5*self.params["zeta"]*(self.E-tmp)**2
        # lb = term1 - term2
        # lb[mask] = 0.

        # Compute Evidence Lower Bound using the lower bound to the likelihood

        # calculate E(Z)E(W)
        ZW = Z.dot(SW.T)
        ZW[mask] = 0.

        # Calculate E[(ZW_nd)^2]
        # this is equal to E[\sum_{k != k} z_k w_k z_k' w_k'] + E[\sum_{k} z_k^2 w_k^2]
        tmp1 = s.square(ZW) - s.dot(s.square(Z),s.square(SW).T) # this is for terms in k != k'
        tmp2 = ZZ.dot(SWW.T) # this is for terms in k = k'
        EZZWW = tmp1 + tmp2

        # calculate elbo terms
        term1 = 0.5 * ((2.*self.obs.data - 1.)*ZW - zeta)
        term2 = - s.log(1 + s.exp(-zeta))
        term3 = - 1/(4 * zeta) *  s.tanh(zeta/2.) * (EZZWW - zeta**2)

        lb = term1 + term2 + term3
        lb[mask] = 0.

        return lb.sum()
